"""
Conformal Models
---------------

A collection of conformal prediction models for pre-trained global forecasting models.
"""

import copy
import math
import os
import sys
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, BinaryIO, Callable, Optional, Union

from darts.utils.likelihood_models.base import (
    Likelihood,
    LikelihoodType,
    quantile_names,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import numpy as np
import pandas as pd

from darts import TimeSeries, metrics
from darts.dataprocessing.pipeline import Pipeline
from darts.dataprocessing.transformers import BaseDataTransformer
from darts.logging import get_logger, raise_log
from darts.metrics.metrics import METRIC_TYPE
from darts.models.forecasting.forecasting_model import GlobalForecastingModel
from darts.utils import _build_tqdm_iterator, _with_sanity_checks
from darts.utils.historical_forecasts.utils import (
    _adjust_historical_forecasts_time_index,
)
from darts.utils.timeseries_generation import _build_forecast_series
from darts.utils.ts_utils import (
    SeriesType,
    get_series_seq_type,
    series2seq,
)
from darts.utils.utils import (
    TORCH_AVAILABLE,
    _check_quantiles,
    generate_index,
    n_steps_between,
    random_method,
    sample_from_quantiles,
)

if TORCH_AVAILABLE:
    from darts.models.forecasting.torch_forecasting_model import TorchForecastingModel
else:
    TorchForecastingModel = None

logger = get_logger(__name__)


class ConformalModel(GlobalForecastingModel, ABC):
    @random_method
    def __init__(
        self,
        model: GlobalForecastingModel,
        quantiles: list[float],
        symmetric: bool = True,
        cal_length: Optional[int] = None,
        cal_stride: int = 1,
        cal_num_samples: int = 500,
        random_state: Optional[int] = None,
    ):
        """Base Conformal Prediction Model.

        Base class for any conformal prediction model. A conformal model calibrates the predictions from any
        pre-trained global forecasting model. It does not have to be trained, and can generate calibrated forecasts
        directly using the underlying trained forecasting model. Since it is a probabilistic model, you can generate
        forecasts in two ways (when calling `predict()`, `historical_forecasts()`, ...):

        - Predict the calibrated quantile intervals directly: Pass parameters `predict_likelihood_parameters=True`, and
          `num_samples=1` to the forecast method.
        - Predict stochastic samples from the calibrated quantile intervals: Pass parameters
          `predict_likelihood_parameters=False`, and `num_samples>>1` to the forecast method.

        Conformal models can be applied to any of Darts' global forecasting model, as long as the model has been
        fitted before. In general the workflow of the models to produce one calibrated forecast/prediction is as
        follows:

        - Extract a calibration set: The calibration set for each conformal forecast is automatically extracted from
          the most recent past of your input series relative to the forecast start point. The number of calibration
          examples (forecast errors / non-conformity scores) to consider can be defined at model creation with
          parameter `cal_length`. Note that when using `cal_stride>1`, a longer history is required since the
          calibration examples are generated with stridden historical forecasts.
        - Generate historical forecasts on the calibration set (using the forecasting model) with a stride `cal_stride`.
        - Compute the errors/non-conformity scores (specific to each conformal model) on these historical forecasts
        - Compute the quantile values from the errors / non-conformity scores (using our desired quantiles set at model
          creation with parameter `quantiles`).
        - Compute the conformal prediction: Using these quantile values, add calibrated intervals to (or adjust the
          existing intervals of) the forecasting model's predictions.

        Some notes:

        - When computing `historical_forecasts()`, `backtest()`, `residuals()`, ... the above is applied for each
          forecast (the forecasting model's historical forecasts are only generated once for efficiency).
        - For multi-horizon forecasts, the above is applied for each step in the horizon separately.

        Parameters
        ----------
        model
            A pre-trained global forecasting model. See the list of models
            `here <https://unit8co.github.io/darts/#forecasting-models>`_.
        quantiles
            A list of quantiles centered around the median `q=0.5` to use. For example quantiles
            [0.1, 0.2, 0.5, 0.8 0.9] correspond to two intervals with (0.9 - 0.1) = 80%, and (0.8 - 0.2) 60% coverage
            around the median (model forecast).
        symmetric
            Whether to use symmetric non-conformity scores. If `False`, uses asymmetric scores (individual scores
            for lower- and upper quantile interval bounds).
        cal_length
            The number of past forecast errors / non-conformity scores to use as calibration for each conformal
            forecast (and each step in the horizon). If `None`, considers all scores.
        cal_stride
            The stride to apply when computing the historical forecasts and non-conformity scores on the calibration
            set. The actual conformal forecasts can have a different stride given with parameter `stride` in downstream
            tasks (e.g. historical forecasts, backtest, ...)
        cal_num_samples
            The number of samples to generate for each calibration forecast (if `model` is a probabilistic forecasting
            model). The non-conformity scores are computed on the quantile values of these forecasts (using quantiles
            `quantiles`). Uses `1` for deterministic models. The actual conformal forecasts can have a different number
            of samples given with parameter `num_samples` in downstream tasks (e.g. predict, historical forecasts, ...).
        random_state
            Control the randomness of probabilistic conformal forecasts (sample generation) across different runs.
        """
        if not isinstance(model, GlobalForecastingModel) or not model._fit_called:
            raise_log(
                ValueError("`model` must be a pre-trained `GlobalForecastingModel`."),
                logger=logger,
            )
        _check_quantiles(quantiles)

        if cal_length is not None and cal_length < 1:
            raise_log(
                ValueError("`cal_length` must be `>=1` or `None`."), logger=logger
            )
        if cal_stride < 1:
            raise_log(ValueError("`cal_stride` must be `>=1`."), logger=logger)
        if cal_num_samples < 1:
            raise_log(ValueError("`cal_num_samples` must be `>=1`."), logger=logger)

        super().__init__(add_encoders=None)

        # quantiles and interval setup
        self.quantiles = np.array(quantiles)
        self.idx_median = quantiles.index(0.5)
        self.q_interval = [
            (q_l, q_h)
            for q_l, q_h in zip(
                quantiles[: self.idx_median], quantiles[self.idx_median + 1 :][::-1]
            )
        ]
        self.interval_range = np.array([
            q_high - q_low for q_low, q_high in self.q_interval
        ])

        if symmetric:
            # symmetric considers both tails together
            self.interval_range_sym = copy.deepcopy(self.interval_range)
        else:
            # asymmetric considers tails separately
            self.interval_range_sym = 1 - (1 - self.interval_range) / 2
        self.symmetric = symmetric

        # model setup
        self.model = model
        self.cal_length = cal_length
        self.cal_stride = cal_stride
        self.cal_num_samples = (
            cal_num_samples if model.supports_probabilistic_prediction else 1
        )
        self._likelihood = Likelihood(
            likelihood_type=LikelihoodType.Quantile,
            parameter_names=quantile_names(quantiles),
        )
        self._fit_called = True

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        **kwargs,
    ) -> "ConformalModel":
        """Fit/train the underlying forecasting model on (potentially multiple) series.

        Optionally, one or multiple past and/or future covariates series can be provided as well, depending on the
        forecasting model used. The number of covariates series must match the number of target series.

        Notes
        -----
        Conformal Models do not require calling `fit()`, since they use pre-trained global forecasting models.
        You can call `predict()` directly. Also, make sure that the input series used in `predict()` corresponds to
        a calibration set, and not the same as used during training with `fit()`.

        Parameters
        ----------
        series
            One or several target time series. The model will be trained to forecast these time series.
            The series may or may not be multivariate, but if multiple series are provided
            they must have the same number of components.
        past_covariates
            One or several past-observed covariate time series. These time series will not be forecast, but can
            be used by some models as an input. The covariate(s) may or may not be multivariate, but if multiple
            covariates are provided they must have the same number of components. If `past_covariates` is provided,
            it must contain the same number of series as `series`.
        future_covariates
            One or several future-known covariate time series. These time series will not be forecast, but can
            be used by some models as an input. The covariate(s) may or may not be multivariate, but if multiple
            covariates are provided they must have the same number of components. If `future_covariates` is provided,
            it must contain the same number of series as `series`.
        **kwargs
            Optional keyword arguments that will passed to the underlying forecasting model's `fit()` method.

        Returns
        -------
        self
            Fitted model.
        """
        # does not have to be trained, but we allow it for unified API
        self.model.fit(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            **kwargs,
        )
        return self

    def predict(
        self,
        n: int,
        series: Union[TimeSeries, Sequence[TimeSeries]] = None,
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        verbose: bool = False,
        predict_likelihood_parameters: bool = False,
        show_warnings: bool = True,
        **kwargs,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Forecasts calibrated quantile intervals (or samples from calibrated intervals) for `n` time steps after the
        end of the `series`.

        It is important that the input series for prediction correspond to a calibration set - a set different to the
        series that the underlying forecasting `model` was trained on.

        Since it is a probabilistic model, you can generate forecasts in two ways:

        - Predict the calibrated quantile intervals directly: Pass parameters `predict_likelihood_parameters=True`, and
          `num_samples=1` to the forecast method.
        - Predict stochastic samples from the calibrated quantile intervals: Pass parameters
          `predict_likelihood_parameters=False`, and `num_samples>>1` to the forecast method.

        Under the hood, the simplified workflow to produce one calibrated forecast/prediction for every step in the
        horizon `n` is as follows (note: `cal_length` and `cal_stride` can be set at model creation):

        - Extract a calibration set: The calibration set for each conformal forecast is automatically extracted from
          the most recent past of your input series relative to the forecast start point. The number of calibration
          examples (forecast errors / non-conformity scores) to consider can be defined at model creation
          with parameter `cal_length`. Note that when using `cal_stride>1`, a longer history is required since
          the calibration examples are generated with stridden historical forecasts.
        - Generate historical forecasts on the calibration set (using the forecasting model) with a stride `cal_stride`.
        - Compute the errors/non-conformity scores (specific to each conformal model) on these historical forecasts
        - Compute the quantile values from the errors / non-conformity scores (using our desired quantiles set at model
          creation with parameter `quantiles`).
        - Compute the conformal prediction: Using these quantile values, add calibrated intervals to (or adjust the
          existing intervals of) the forecasting model's predictions.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        series
            A series or sequence of series, representing the history of the target series whose future is to be
            predicted. Will use the past of this series for calibration. The series should not have any overlap with
            the series used to train the forecasting model.
        past_covariates
            Optionally, a (sequence of) past-observed covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        future_covariates
            Optionally, a (sequence of) future-known covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        num_samples
            Number of times a prediction is sampled from the calibrated quantile predictions using linear
            interpolation in-between the quantiles. For larger values, the sample distribution approximates the
            calibrated quantile predictions.
        verbose
            Whether to print the progress.
        predict_likelihood_parameters
            If set to `True`, generates the quantile predictions directly. Only supported with `num_samples = 1`.
        show_warnings
            Whether to show warnings related auto-regression and past covariates usage.
        **kwargs
            Optional keyword arguments that will passed to the underlying forecasting model's `predict()` and
            `historical_forecasts()` methods.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            If `series` is not specified, this function returns a single time series containing the `n`
            next points after then end of the training series.
            If `series` is given and is a simple ``TimeSeries``, this function returns the `n` next points
            after the end of `series`.
            If `series` is given and is a sequence of several time series, this function returns
            a sequence where each element contains the corresponding `n` points forecasts.
        """
        # call predict to verify that all series have required input times
        _ = self.model.predict(
            n=n,
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            num_samples=self.cal_num_samples,
            verbose=verbose,
            predict_likelihood_parameters=False,
            show_warnings=show_warnings,
            **kwargs,
        )

        series = series or self.model.training_series
        called_with_single_series = get_series_seq_type(series) == SeriesType.SINGLE
        series = series2seq(series)

        # generate only the required forecasts for calibration (including the last forecast which is the output of
        # `predict()`)
        cal_start, cal_start_format = _get_calibration_hfc_start(
            series=series,
            horizon=n,
            output_chunk_shift=self.output_chunk_shift,
            cal_length=self.cal_length,
            cal_stride=self.cal_stride,
            start="end",
            start_format="position",
        )

        cal_hfcs = self.model.historical_forecasts(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=n,
            num_samples=self.cal_num_samples,
            start=cal_start,
            start_format=cal_start_format,
            stride=self.cal_stride,
            retrain=False,
            overlap_end=True,
            last_points_only=False,
            verbose=verbose,
            show_warnings=False,
            predict_likelihood_parameters=False,
            predict_kwargs=kwargs,
        )
        cal_preds = self._calibrate_forecasts(
            series=series,
            forecasts=cal_hfcs,
            num_samples=num_samples,
            start="end",  # uses last hist fc (output of `predict()`)
            start_format="position",
            forecast_horizon=n,
            stride=self.cal_stride,
            overlap_end=True,
            last_points_only=False,
            verbose=verbose,
            show_warnings=show_warnings,
            predict_likelihood_parameters=predict_likelihood_parameters,
        )
        # convert historical forecasts output to simple forecast / prediction
        if called_with_single_series:
            return cal_preds[0][0]
        else:
            return [cp[0] for cp in cal_preds]

    @_with_sanity_checks("_historical_forecasts_sanity_checks")
    def historical_forecasts(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        forecast_horizon: int = 1,
        num_samples: int = 1,
        train_length: Optional[int] = None,
        start: Optional[Union[pd.Timestamp, int]] = None,
        start_format: Literal["position", "value"] = "value",
        stride: int = 1,
        retrain: Union[bool, int, Callable[..., bool]] = True,
        overlap_end: bool = False,
        last_points_only: bool = True,
        verbose: bool = False,
        show_warnings: bool = True,
        predict_likelihood_parameters: bool = False,
        enable_optimization: bool = True,
        data_transformers: Optional[
            dict[str, Union[BaseDataTransformer, Pipeline]]
        ] = None,
        fit_kwargs: Optional[dict[str, Any]] = None,
        predict_kwargs: Optional[dict[str, Any]] = None,
        sample_weight: Optional[Union[TimeSeries, Sequence[TimeSeries], str]] = None,
    ) -> Union[TimeSeries, list[TimeSeries], list[list[TimeSeries]]]:
        """Generates calibrated historical forecasts by simulating predictions at various points in time throughout the
        history of the provided (potentially multiple) `series`. This process involves retrospectively applying the
        model to different time steps, as if the forecasts were made in real-time at those specific moments. This
        allows for an evaluation of the model's performance over the entire duration of the series, providing insights
        into its predictive accuracy and robustness across different historical periods.

        Currently, conformal models only support the pre-trained historical forecasts mode (`retrain=False`).
        Parameters `retrain` and `train_length` are ignored.

        **Pre-trained Mode:** First, all historical forecasts are generated using the underlying pre-trained global
        forecasting model (see :meth:`ForecastingModel.historical_forecasts()
        <darts.models.forecasting.forecasting_model.ForecastingModel.historical_forecasts>` for more info). Then it
        repeatedly builds a calibration set by either expanding from the beginning of the historical forecasts or by
        using a fixed-length moving window with length `cal_length` (the start point can also be configured with
        `start` and `start_format`).
        The next forecast of length `forecast_horizon` is then calibrated on this calibration set. Subsequently, the
        end of the calibration set is moved forward by `stride` time steps, and the process is repeated.

        By default, with `last_points_only=True`, this method returns a single time series (or a sequence of time
        series when `series` is also a sequence of series) composed of the last point from each calibrated historical
        forecast. This time series will thus have a frequency of `series.freq * stride`.
        If `last_points_only=False`, it will instead return a list (or a sequence of lists) with all calibrated
        historical forecasts of length `forecast_horizon` and frequency `series.freq`.

        Parameters
        ----------
        series
            A (sequence of) target time series used to successively compute the historical forecasts. Will use the past
            of this series for calibration. The series should not have any overlap with the series used to train the
            forecasting model.
        past_covariates
            Optionally, a (sequence of) past-observed covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        future_covariates
            Optionally, a (sequence of) future-known covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        forecast_horizon
            The forecast horizon for the predictions.
        num_samples
            Number of times a prediction is sampled from the calibrated quantile predictions using linear
            interpolation in-between the quantiles. For larger values, the sample distribution approximates the
            calibrated quantile predictions.
        train_length
            Currently ignored by conformal models.
        start
            Optionally, the first point in time at which a prediction is computed. This parameter supports:
            ``int``, ``pandas.Timestamp``, and ``None``.
            If an ``int``, it is either the index position of the first prediction point for `series` with a
            `pd.DatetimeIndex`, or the index value for `series` with a `pd.RangeIndex`. The latter can be changed to
            the index position with `start_format="position"`.
            If a ``pandas.Timestamp``, it is the time stamp of the first prediction point.
            If ``None``, the first prediction point will automatically be set to:

            - the first predictable point if `retrain` is ``False``, or `retrain` is a Callable and the first
              predictable point is earlier than the first trainable point.
            - the first trainable point if `retrain` is ``True`` or ``int`` (given `train_length`),
              or `retrain` is a ``Callable`` and the first trainable point is earlier than the first predictable point.
            - the first trainable point (given `train_length`) otherwise

            Note: If the model uses a shifted output (`output_chunk_shift > 0`), then the first predicted point is also
            shifted by `output_chunk_shift` points into the future.
            Note: Raises a ValueError if `start` yields a time outside the time index of `series`.
            Note: If `start` is outside the possible historical forecasting times, will ignore the parameter
            (default behavior with ``None``) and start at the first trainable/predictable point.
        start_format
            Defines the `start` format.
            If set to ``'position'``, `start` corresponds to the index position of the first predicted point and can
            range from `(-len(series), len(series) - 1)`.
            If set to ``'value'``, `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: ``'value'``.
        stride
            The number of time steps between two consecutive predictions. Must be a round-multiple of `cal_stride`
            (set at model creation) and `>=cal_stride`.
        retrain
            Currently ignored by conformal models.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not.
        last_points_only
            Whether to return only the last point of each historical forecast. If set to ``True``, the method returns a
            single ``TimeSeries`` (for each time series in `series`) containing the successive point forecasts.
            Otherwise, returns a list of historical ``TimeSeries`` forecasts.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        predict_likelihood_parameters
            If set to `True`, generates the quantile predictions directly. Only supported with `num_samples = 1`.
        enable_optimization
            Whether to use the optimized version of `historical_forecasts` when supported and available.
            Default: ``True``.
        data_transformers
            Optionally, a dictionary of `BaseDataTransformer` or `Pipeline` to apply to the corresponding series
            (possibles keys; "series", "past_covariates", "future_covariates"). If provided, all input series must be
            in the un-transformed space. For fittable transformer / pipeline:

            - if `retrain=True`, the data transformer re-fit on the training data at each historical forecast step
              (currently ignored by conformal models).
            - if `retrain=False`, the data transformer transforms the series once before all the forecasts.

            The fitted transformer is used to transform the input during both training and prediction.
            If the transformation is invertible, the forecasts will be inverse-transformed.
        fit_kwargs
            Currently ignored by conformal models.
        predict_kwargs
            Optionally, some additional arguments passed to the model `predict()` method.
        sample_weight
            Currently ignored by conformal models.

        Returns
        -------
        TimeSeries
            A single historical forecast for a single `series` and `last_points_only=True`: it contains only the
            predictions at step `forecast_horizon` from all historical forecasts.
        list[TimeSeries]
            A list of historical forecasts for:

            - a sequence (list) of `series` and `last_points_only=True`: for each series, it contains only the
              predictions at step `forecast_horizon` from all historical forecasts.
            - a single `series` and `last_points_only=False`: for each historical forecast, it contains the entire
              horizon `forecast_horizon`.
        list[list[TimeSeries]]
            A list of lists of historical forecasts for a sequence of `series` and `last_points_only=False`. For each
            series, and historical forecast, it contains the entire horizon `forecast_horizon`. The outer list
            is over the series provided in the input sequence, and the inner lists contain the historical forecasts for
            each series.
        """
        called_with_single_series = get_series_seq_type(series) == SeriesType.SINGLE
        series = series2seq(series)
        past_covariates = series2seq(past_covariates)
        future_covariates = series2seq(future_covariates)

        # generate only the required forecasts (if `start` is given, we have to start earlier to satisfy the
        # calibration set requirements)
        cal_start, cal_start_format = _get_calibration_hfc_start(
            series=series,
            horizon=forecast_horizon,
            output_chunk_shift=self.output_chunk_shift,
            cal_length=self.cal_length,
            cal_stride=self.cal_stride,
            start=start,
            start_format=start_format,
        )
        hfcs = self.model.historical_forecasts(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=forecast_horizon,
            num_samples=self.cal_num_samples,
            start=cal_start,
            start_format=cal_start_format,
            stride=self.cal_stride,
            retrain=False,
            overlap_end=overlap_end,
            last_points_only=last_points_only,
            verbose=verbose,
            show_warnings=False,
            predict_likelihood_parameters=False,
            enable_optimization=enable_optimization,
            data_transformers=data_transformers,
            fit_kwargs=fit_kwargs,
            predict_kwargs=predict_kwargs,
        )
        calibrated_forecasts = self._calibrate_forecasts(
            series=series,
            forecasts=hfcs,
            num_samples=num_samples,
            start=start,
            start_format=start_format,
            forecast_horizon=forecast_horizon,
            stride=stride,
            overlap_end=overlap_end,
            last_points_only=last_points_only,
            verbose=verbose,
            show_warnings=show_warnings,
            predict_likelihood_parameters=predict_likelihood_parameters,
        )
        return (
            calibrated_forecasts[0]
            if called_with_single_series
            else calibrated_forecasts
        )

    def backtest(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        historical_forecasts: Optional[
            Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
        ] = None,
        forecast_horizon: int = 1,
        num_samples: int = 1,
        train_length: Optional[int] = None,
        start: Optional[Union[pd.Timestamp, int]] = None,
        start_format: Literal["position", "value"] = "value",
        stride: int = 1,
        retrain: Union[bool, int, Callable[..., bool]] = True,
        overlap_end: bool = False,
        last_points_only: bool = False,
        metric: Union[METRIC_TYPE, list[METRIC_TYPE]] = metrics.mape,
        reduction: Union[Callable[..., float], None] = np.mean,
        verbose: bool = False,
        show_warnings: bool = True,
        predict_likelihood_parameters: bool = False,
        enable_optimization: bool = True,
        data_transformers: Optional[
            dict[str, Union[BaseDataTransformer, Pipeline]]
        ] = None,
        metric_kwargs: Optional[Union[dict[str, Any], list[dict[str, Any]]]] = None,
        fit_kwargs: Optional[dict[str, Any]] = None,
        predict_kwargs: Optional[dict[str, Any]] = None,
        sample_weight: Optional[Union[TimeSeries, Sequence[TimeSeries], str]] = None,
    ) -> Union[float, np.ndarray, list[float], list[np.ndarray]]:
        """Compute error values that the model produced for historical forecasts on (potentially multiple) `series`.

        If `historical_forecasts` are provided, the metric(s) (given by the `metric` function) is evaluated directly on
        all forecasts and actual values. The same `series` and `last_points_only` value must be passed that were used
        to generate the historical forecasts. Finally, the method returns an optional `reduction` (the mean by default)
        of all these metric scores.

        If `historical_forecasts` is ``None``, it first generates the historical forecasts with the parameters given
        below (see :meth:`ConformalModel.historical_forecasts()
        <darts.models.forecasting.conformal_models.ConformalModel.historical_forecasts>` for more info) and then
        evaluates as described above.

        The metric(s) can be further customized `metric_kwargs` (e.g. control the aggregation over components, time
        steps, multiple series, other required arguments such as `q` for quantile metrics, ...).

        Notes
        -----
        Darts has several metrics to evaluate probabilistic forecasts. For conformal models, we recommend using
        quantile interval metrics (see `here <https://unit8co.github.io/darts/generated_api/darts.metrics.html>`_).
        You can specify which intervals to evaluate by setting `metric_kwargs={'q_interval': my_intervals}`. To check
        all intervals used by your conformal model `my_model`, you can set ``{'q_interval': my_model.q_interval}``.

        Parameters
        ----------
        series
            A (sequence of) target time series used to successively compute the historical forecasts. Will use the past
            of this series for calibration. The series should not have any overlap with the series used to train the
            forecasting model.
        past_covariates
            Optionally, a (sequence of) past-observed covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        future_covariates
            Optionally, a (sequence of) future-known covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        historical_forecasts
            Optionally, the (or a sequence of / a sequence of sequences of) historical forecasts time series to be
            evaluated. Corresponds to the output of :meth:`historical_forecasts()
            <darts.models.forecasting.conformal_models.ConformalModel.historical_forecasts>`. The same `series` and
            `last_points_only` values must be passed that were used to generate the historical forecasts. If provided,
            will skip historical forecasting and ignore all parameters except `series`, `last_points_only`, `metric`,
            and `reduction`.
        forecast_horizon
            The forecast horizon for the predictions.
        num_samples
            Number of times a prediction is sampled from the calibrated quantile predictions using linear
            interpolation in-between the quantiles. For larger values, the sample distribution approximates the
            calibrated quantile predictions.
        train_length
            Currently ignored by conformal models.
        start
            Optionally, the first point in time at which a prediction is computed. This parameter supports:
            ``int``, ``pandas.Timestamp``, and ``None``.
            If an ``int``, it is either the index position of the first prediction point for `series` with a
            `pd.DatetimeIndex`, or the index value for `series` with a `pd.RangeIndex`. The latter can be changed to
            the index position with `start_format="position"`.
            If a ``pandas.Timestamp``, it is the time stamp of the first prediction point.
            If ``None``, the first prediction point will automatically be set to:

            - the first predictable point if `retrain` is ``False``, or `retrain` is a Callable and the first
              predictable point is earlier than the first trainable point.
            - the first trainable point if `retrain` is ``True`` or ``int`` (given `train_length`),
              or `retrain` is a ``Callable`` and the first trainable point is earlier than the first predictable point.
            - the first trainable point (given `train_length`) otherwise

            Note: If the model uses a shifted output (`output_chunk_shift > 0`), then the first predicted point is also
            shifted by `output_chunk_shift` points into the future.
            Note: Raises a ValueError if `start` yields a time outside the time index of `series`.
            Note: If `start` is outside the possible historical forecasting times, will ignore the parameter
            (default behavior with ``None``) and start at the first trainable/predictable point.
        start_format
            Defines the `start` format.
            If set to ``'position'``, `start` corresponds to the index position of the first predicted point and can
            range from `(-len(series), len(series) - 1)`.
            If set to ``'value'``, `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: ``'value'``.
        stride
            The number of time steps between two consecutive predictions.
        retrain
            Currently ignored by conformal models.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not.
        last_points_only
            Whether to return only the last point of each historical forecast. If set to ``True``, the method returns a
            single ``TimeSeries`` (for each time series in `series`) containing the successive point forecasts.
            Otherwise, returns a list of historical ``TimeSeries`` forecasts.
        metric
            A metric function or a list of metric functions. Each metric must either be a Darts metric (see `here
            <https://unit8co.github.io/darts/generated_api/darts.metrics.html>`_), or a custom metric that has an
            identical signature as Darts' metrics, uses decorators :func:`~darts.metrics.metrics.multi_ts_support` and
            :func:`~darts.metrics.metrics.multi_ts_support`, and returns the metric score.
        reduction
            A function used to combine the individual error scores obtained when `last_points_only` is set to `False`.
            When providing several metric functions, the function will receive the argument `axis = 1` to obtain single
            value for each metric function.
            If explicitly set to `None`, the method will return a list of the individual error scores instead.
            Set to ``np.mean`` by default.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        predict_likelihood_parameters
            If set to `True`, generates the quantile predictions directly. Only supported with `num_samples = 1`.
        enable_optimization
            Whether to use the optimized version of `historical_forecasts` when supported and available.
            Default: ``True``.
        data_transformers
            Optionally, a dictionary of `BaseDataTransformer` or `Pipeline` to apply to the corresponding series
            (possibles keys; "series", "past_covariates", "future_covariates"). If provided, all input series must be
            in the un-transformed space. For fittable transformer / pipeline:

            - if `retrain=True`, the data transformer re-fit on the training data at each historical forecast step
              (currently ignored by conformal models).
            - if `retrain=False`, the data transformer transforms the series once before all the forecasts.

            The fitted transformer is used to transform the input during both training and prediction.
            If the transformation is invertible, the forecasts will be inverse-transformed.
            Only effective when `historical_forecasts=None`.
        metric_kwargs
            Additional arguments passed to `metric()`, such as `'n_jobs'` for parallelization, `'component_reduction'`
            for reducing the component wise metrics, seasonality `'m'` for scaled metrics, etc. Will pass arguments to
            each metric separately and only if they are present in the corresponding metric signature. Parameter
            `'insample'` for scaled metrics (e.g. mase`, `rmsse`, ...) is ignored, as it is handled internally.
        fit_kwargs
            Currently ignored by conformal models.
        predict_kwargs
            Optionally, some additional arguments passed to the model `predict()` method.
        sample_weight
            Currently ignored by conformal models.

        Returns
        -------
        float
            A single backtest score for single uni/multivariate series, a single `metric` function and:

            - `historical_forecasts` generated with `last_points_only=True`
            - `historical_forecasts` generated with `last_points_only=False` and using a backtest `reduction`
        np.ndarray
            An numpy array of backtest scores. For single series and one of:

            - a single `metric` function, `historical_forecasts` generated with `last_points_only=False`
              and backtest `reduction=None`. The output has shape (n forecasts, *).
            - multiple `metric` functions and `historical_forecasts` generated with `last_points_only=False`.
              The output has shape (*, n metrics) when using a backtest `reduction`, and (n forecasts, *, n metrics)
              when `reduction=None`
            - multiple uni/multivariate series including `series_reduction` and at least one of
              `component_reduction=None` or `time_reduction=None` for "per time step metrics"
        list[float]
            Same as for type `float` but for a sequence of series. The returned metric list has length
            `len(series)` with the `float` metric for each input `series`.
        list[np.ndarray]
            Same as for type `np.ndarray` but for a sequence of series. The returned metric list has length
            `len(series)` with the `np.ndarray` metrics for each input `series`.
        """
        return super().backtest(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            historical_forecasts=historical_forecasts,
            forecast_horizon=forecast_horizon,
            num_samples=num_samples,
            train_length=train_length,
            start=start,
            start_format=start_format,
            stride=stride,
            retrain=retrain,
            overlap_end=overlap_end,
            last_points_only=last_points_only,
            metric=metric,
            reduction=reduction,
            verbose=verbose,
            show_warnings=show_warnings,
            predict_likelihood_parameters=predict_likelihood_parameters,
            enable_optimization=enable_optimization,
            data_transformers=data_transformers,
            metric_kwargs=metric_kwargs,
            fit_kwargs=fit_kwargs,
            predict_kwargs=predict_kwargs,
            sample_weight=sample_weight,
        )

    def residuals(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        historical_forecasts: Optional[
            Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
        ] = None,
        forecast_horizon: int = 1,
        num_samples: int = 1,
        train_length: Optional[int] = None,
        start: Optional[Union[pd.Timestamp, int]] = None,
        start_format: Literal["position", "value"] = "value",
        stride: int = 1,
        retrain: Union[bool, int, Callable[..., bool]] = True,
        overlap_end: bool = False,
        last_points_only: bool = True,
        metric: METRIC_TYPE = metrics.err,
        verbose: bool = False,
        show_warnings: bool = True,
        predict_likelihood_parameters: bool = False,
        enable_optimization: bool = True,
        data_transformers: Optional[
            dict[str, Union[BaseDataTransformer, Pipeline]]
        ] = None,
        metric_kwargs: Optional[dict[str, Any]] = None,
        fit_kwargs: Optional[dict[str, Any]] = None,
        predict_kwargs: Optional[dict[str, Any]] = None,
        sample_weight: Optional[Union[TimeSeries, Sequence[TimeSeries], str]] = None,
        values_only: bool = False,
    ) -> Union[TimeSeries, list[TimeSeries], list[list[TimeSeries]]]:
        """Compute the residuals that the model produced for historical forecasts on (potentially multiple) `series`.

        This function computes the difference (or one of Darts' "per time step" metrics) between the actual
        observations from `series` and the fitted values obtained by training the model on `series` (or using a
        pre-trained model with `retrain=False`). Not all models support fitted values, so we use historical forecasts
        as an approximation for them.

        In sequence this method performs:

        - use pre-computed `historical_forecasts` or compute historical forecasts for each series (see
          :meth:`~darts.models.forecasting.conformal_models.ConformalModel.historical_forecasts` for more details).
          How the historical forecasts are generated can be configured with parameters `num_samples`, `train_length`,
          `start`, `start_format`, `forecast_horizon`, `stride`, `retrain`, `last_points_only`, `fit_kwargs`, and
          `predict_kwargs`.
        - compute a backtest using a "per time step" `metric` between the historical forecasts and `series` per
          component/column and time step (see
          :meth:`~darts.models.forecasting.conformal_models.ConformalModel.backtest` for more details). By default,
          uses the residuals :func:`~darts.metrics.metrics.err` (error) as a `metric`.
        - create and return `TimeSeries` (or simply a np.ndarray with `values_only=True`) with the time index from
          historical forecasts, and values from the metrics per component and time step.

        This method works for single or multiple univariate or multivariate series.
        It uses the median prediction (when dealing with stochastic forecasts).

        Notes
        -----
        Darts has several metrics to evaluate probabilistic forecasts. For conformal models, we recommend using
        "per time step" quantile interval metrics (see `here
        <https://unit8co.github.io/darts/generated_api/darts.metrics.html>`_). You can specify which intervals to
        evaluate by setting `metric_kwargs={'q_interval': my_intervals}`. To check all intervals used by your conformal
        model `my_model`, you can set ``{'q_interval': my_model.q_interval}``.

        Parameters
        ----------
        series
            A (sequence of) target time series used to successively compute the historical forecasts. Will use the past
            of this series for calibration. The series should not have any overlap with the series used to train the
            forecasting model.
        past_covariates
            Optionally, a (sequence of) past-observed covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        future_covariates
            Optionally, a (sequence of) future-known covariate time series for every input time series in `series`.
            Their dimension must match that of the past covariates used for training. Will use this series for
            calibration.
        historical_forecasts
            Optionally, the (or a sequence of / a sequence of sequences of) historical forecasts time series to be
            evaluated. Corresponds to the output of :meth:`historical_forecasts()
            <darts.models.forecasting.conformal_models.ConformalModel.historical_forecasts>`. The same `series` and
            `last_points_only` values must be passed that were used to generate the historical forecasts. If provided,
            will skip historical forecasting and ignore all parameters except `series`, `last_points_only`, `metric`,
            and `reduction`.
        forecast_horizon
            The forecast horizon for the predictions.
        num_samples
            Number of times a prediction is sampled from the calibrated quantile predictions using linear
            interpolation in-between the quantiles. For larger values, the sample distribution approximates the
            calibrated quantile predictions.
        train_length
            Currently ignored by conformal models.
        start
            Optionally, the first point in time at which a prediction is computed. This parameter supports:
            ``int``, ``pandas.Timestamp``, and ``None``.
            If an ``int``, it is either the index position of the first prediction point for `series` with a
            `pd.DatetimeIndex`, or the index value for `series` with a `pd.RangeIndex`. The latter can be changed to
            the index position with `start_format="position"`.
            If a ``pandas.Timestamp``, it is the time stamp of the first prediction point.
            If ``None``, the first prediction point will automatically be set to:

            - the first predictable point if `retrain` is ``False``, or `retrain` is a Callable and the first
              predictable point is earlier than the first trainable point.
            - the first trainable point if `retrain` is ``True`` or ``int`` (given `train_length`),
              or `retrain` is a ``Callable`` and the first trainable point is earlier than the first predictable point.
            - the first trainable point (given `train_length`) otherwise

            Note: If the model uses a shifted output (`output_chunk_shift > 0`), then the first predicted point is also
            shifted by `output_chunk_shift` points into the future.
            Note: Raises a ValueError if `start` yields a time outside the time index of `series`.
            Note: If `start` is outside the possible historical forecasting times, will ignore the parameter
            (default behavior with ``None``) and start at the first trainable/predictable point.
        start_format
            Defines the `start` format.
            If set to ``'position'``, `start` corresponds to the index position of the first predicted point and can
            range from `(-len(series), len(series) - 1)`.
            If set to ``'value'``, `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: ``'value'``.
        stride
            The number of time steps between two consecutive predictions.
        retrain
            Currently ignored by conformal models.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not.
        last_points_only
            Whether to return only the last point of each historical forecast. If set to ``True``, the method returns a
            single ``TimeSeries`` (for each time series in `series`) containing the successive point forecasts.
            Otherwise, returns a list of historical ``TimeSeries`` forecasts.
        metric
            Either one of Darts' "per time step" metrics (see `here
            <https://unit8co.github.io/darts/generated_api/darts.metrics.html>`_), or a custom metric that has an
            identical signature as Darts' "per time step" metrics, uses decorators
            :func:`~darts.metrics.metrics.multi_ts_support` and :func:`~darts.metrics.metrics.multi_ts_support`,
            and returns one value per time step.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        predict_likelihood_parameters
            If set to `True`, generates the quantile predictions directly. Only supported with `num_samples = 1`.
        enable_optimization
            Whether to use the optimized version of `historical_forecasts` when supported and available.
            Default: ``True``.
        data_transformers
            Optionally, a dictionary of `BaseDataTransformer` or `Pipeline` to apply to the corresponding series
            (possibles keys; "series", "past_covariates", "future_covariates"). If provided, all input series must be
            in the un-transformed space. For fittable transformer / pipeline:

            - if `retrain=True`, the data transformer re-fit on the training data at each historical forecast step
              (currently ignored by conformal models).
            - if `retrain=False`, the data transformer transforms the series once before all the forecasts.

            The fitted transformer is used to transform the input during both training and prediction.
            If the transformation is invertible, the forecasts will be inverse-transformed.
            Only effective when `historical_forecasts=None`.
        metric_kwargs
            Additional arguments passed to `metric()`, such as `'n_jobs'` for parallelization, `'m'` for scaled
            metrics, etc. Will pass arguments only if they are present in the corresponding metric signature. Ignores
            reduction arguments `"series_reduction", "component_reduction", "time_reduction"`, and parameter
            `'insample'` for scaled metrics (e.g. mase`, `rmsse`, ...), as they are handled internally.
        fit_kwargs
            Currently ignored by conformal models.
        predict_kwargs
            Optionally, some additional arguments passed to the model `predict()` method.
        sample_weight
            Currently ignored by conformal models.
        values_only
            Whether to return the residuals as `np.ndarray`. If `False`, returns residuals as `TimeSeries`.

        Returns
        -------
        TimeSeries
            Residual `TimeSeries` for a single `series` and `historical_forecasts` generated with
            `last_points_only=True`.
        list[TimeSeries]
            A list of residual `TimeSeries` for a sequence (list) of `series` with `last_points_only=True`.
            The residual list has length `len(series)`.
        list[list[TimeSeries]]
            A list of lists of residual `TimeSeries` for a sequence of `series` with `last_points_only=False`.
            The outer residual list has length `len(series)`. The inner lists consist of the residuals from
            all possible series-specific historical forecasts.
        """
        return super().residuals(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            historical_forecasts=historical_forecasts,
            forecast_horizon=forecast_horizon,
            num_samples=num_samples,
            train_length=train_length,
            start=start,
            start_format=start_format,
            stride=stride,
            retrain=retrain,
            overlap_end=overlap_end,
            last_points_only=last_points_only,
            metric=metric,
            verbose=verbose,
            show_warnings=show_warnings,
            predict_likelihood_parameters=predict_likelihood_parameters,
            enable_optimization=enable_optimization,
            data_transformers=data_transformers,
            metric_kwargs=metric_kwargs,
            fit_kwargs=fit_kwargs,
            predict_kwargs=predict_kwargs,
            sample_weight=sample_weight,
            values_only=values_only,
        )

    @random_method
    def _calibrate_forecasts(
        self,
        series: Sequence[TimeSeries],
        forecasts: Union[Sequence[Sequence[TimeSeries]], Sequence[TimeSeries]],
        num_samples: int = 1,
        start: Optional[Union[pd.Timestamp, int, str]] = None,
        start_format: Literal["position", "value"] = "value",
        forecast_horizon: int = 1,
        stride: int = 1,
        overlap_end: bool = False,
        last_points_only: bool = True,
        verbose: bool = False,
        show_warnings: bool = True,
        predict_likelihood_parameters: bool = False,
    ) -> Union[TimeSeries, list[TimeSeries], list[list[TimeSeries]]]:
        """Generate calibrated historical forecasts.

        In general the workflow of the models to produce one calibrated forecast/prediction per step in the horizon
        is as follows:

        - Generate historical forecasts for `series` with stride `cal_stride` (using the forecasting model)
        - Extract a calibration set: The forecasts from the most recent past to use as calibration for one conformal
          prediction. The number of examples to use can be defined at model creation with parameter `cal_length`. It
          automatically extracts the calibration set from the most recent past of your input series (`series`,
          `past_covariates`, ...).
        - Compute the errors/non-conformity scores (specific to each conformal model) on these historical forecasts
        - Compute the quantile values from the errors / non-conformity scores (using our desired quantiles set at model
          creation with parameter `quantiles`).
        - Compute the conformal prediction: Using these quantile values, add calibrated intervals to (or adjust the
          existing intervals of) the forecasting model's predictions.
        """
        cal_stride = self.cal_stride
        cal_length = self.cal_length
        metric, metric_kwargs = self._residuals_metric
        residuals = self.model.residuals(
            series=series,
            historical_forecasts=forecasts,
            overlap_end=overlap_end,
            last_points_only=last_points_only,
            verbose=verbose,
            show_warnings=show_warnings,
            values_only=True,
            metric=metric,
            metric_kwargs=metric_kwargs,
        )

        outer_iterator = enumerate(zip(series, forecasts, residuals))
        if len(series) > 1:
            # Use tqdm on the outer loop only if there's more than one series to iterate over
            # (otherwise use tqdm on the inner loop).
            outer_iterator = _build_tqdm_iterator(
                outer_iterator,
                verbose,
                total=len(series),
                desc="conformal forecasts",
            )

        cp_hfcs = []
        for series_idx, (series_, s_hfcs, res) in outer_iterator:
            cp_preds = []

            # no historical forecasts were generated
            if not s_hfcs:
                cp_hfcs.append(cp_preds)
                continue

            last_hfc = s_hfcs if last_points_only else s_hfcs[-1]

            # compute the minimum required number of useful calibration residuals
            # at least one or `cal_length` examples
            min_n_cal = cal_length or 1
            # `last_points_only=False` requires additional examples to use most recent information
            # from all steps in the horizon
            if not last_points_only:
                min_n_cal += math.ceil(forecast_horizon / cal_stride) - 1

            # determine first forecast index for conformal prediction
            # we need at least one residual per point in the horizon prior to the first conformal forecast
            horizon_ocs = forecast_horizon + self.output_chunk_shift
            first_idx_train = math.ceil(horizon_ocs / cal_stride)

            # plus some additional examples based on `cal_length`
            if cal_length is not None:
                first_idx_train += cal_length - 1

            # check if later we need to drop some residuals without useful information (unknown residuals)
            if overlap_end:
                delta_end = n_steps_between(
                    end=last_hfc.end_time(),
                    start=series_.end_time(),
                    freq=series_.freq,
                )
            else:
                delta_end = 0

            # ignore residuals without useful information
            if last_points_only and delta_end > 0:
                # useful residual information only up until the forecast *ending* at the last time step in `series`
                ignore_n_residuals = delta_end
            elif not last_points_only and delta_end >= forecast_horizon:
                # useful residual information only up until the forecast *starting* at the last time step in `series`
                ignore_n_residuals = delta_end - forecast_horizon + 1
            else:
                # ignore at least one forecast residuals from the end, since we can only use prior residuals
                ignore_n_residuals = self.output_chunk_shift + 1
                # with last points only, ignore the last `horizon` residuals to avoid look-ahead bias
                if last_points_only:
                    ignore_n_residuals += forecast_horizon - 1

            # get the last index respecting `cal_stride`
            last_res_idx = -math.ceil(ignore_n_residuals / cal_stride)
            # get only useful residuals
            res = res[:last_res_idx]

            if first_idx_train >= len(s_hfcs) or len(res) < min_n_cal:
                raise_log(
                    ValueError(
                        "Could not build the minimum required calibration input with the provided "
                        f"`series` and `*_covariates` at series index: {series_idx}. "
                        f"Expected to generate at least `{min_n_cal}` calibration forecasts with known residuals "
                        f"before the first conformal forecast, but could only generate `{len(res)}`."
                    ),
                    logger=logger,
                )

            # adjust first index based on `start`
            first_idx_start = 0
            if start == "end":
                # called from `predict()`; start at the last forecast
                first_idx_start = len(s_hfcs) - 1
            elif start is not None:
                # called from `historical_forecasts()`: use user-defined start
                # the conformal forecastable index ranges from the start of the first valid historical
                # forecast until the start of the last historical forecast
                historical_forecasts_time_index = (
                    s_hfcs[first_idx_train].start_time(),
                    s_hfcs[-1].start_time(),
                )
                # adjust forecast start points in case of output shift or `last_points_only=True`
                adjust_idx = (
                    self.output_chunk_shift
                    + int(last_points_only) * (forecast_horizon - 1)
                ) * series_.freq
                historical_forecasts_time_index = (
                    historical_forecasts_time_index[0] - adjust_idx,
                    historical_forecasts_time_index[1] - adjust_idx,
                )

                # adjust forecastable times based on user start, assuming hfcs were generated with `stride=1`
                first_start_time, _ = _adjust_historical_forecasts_time_index(
                    series=series_,
                    series_idx=series_idx,
                    start=start,
                    start_format=start_format,
                    stride=stride,
                    historical_forecasts_time_index=historical_forecasts_time_index,
                    show_warnings=show_warnings,
                )
                # find position relative to start
                first_idx_start = n_steps_between(
                    first_start_time + adjust_idx,
                    s_hfcs[0].start_time(),
                    freq=series_.freq,
                )
                # adjust by stride
                first_idx_start = math.ceil(first_idx_start / cal_stride)

            # get final first index
            first_fc_idx = max([first_idx_train, first_idx_start])
            # bring `res` from shape (forecasting steps, n components, n past residuals) into
            # shape (forecasting steps, n components, n past residuals)
            if last_points_only:
                # -> (1, n components, n samples * n past residuals)
                res = res.transpose(2, 1, 0)
            else:
                # rearrange the residuals to avoid look-ahead bias and to have the same number of examples per
                # point in the horizon. We want the most recent residuals in the past for each step in the horizon.
                res = np.array(res)

                # go through each step in the horizon, use all useful information from the end (most recent values),
                # and skip information at beginning (most distant past);
                # -> (forecast horizon, n components, n past residuals)
                res_ = []
                for idx_horizon in range(forecast_horizon):
                    n = idx_horizon + 1
                    # ignore residuals at beginning
                    idx_fc_start = math.floor((forecast_horizon - n) / cal_stride)
                    # keep as many residuals as possible from end
                    idx_fc_end = -(
                        math.ceil(forecast_horizon / cal_stride) - (idx_fc_start + 1)
                    )
                    res_.append(res[idx_fc_start : idx_fc_end or None, idx_horizon])
                res = np.concatenate(res_, axis=2).T

            # get the last conformal forecast index (exclusive) based on the residual examples
            last_fc_idx = res.shape[2] + math.ceil(horizon_ocs / cal_stride)

            # forecasts are stridden, so stride must be relative
            rel_stride = math.ceil(stride / cal_stride)

            def conformal_predict(idx_, pred_vals_):
                # get the last residual index for calibration, `cal_end` is exclusive
                # to avoid look-ahead bias, use only residuals from before the conformal forecast start point;
                # for `last_points_only=True`, the last residual historically available at the forecasting
                # point is `horizon_ocs - 1` steps before. The same applies to `last_points_only=False` thanks to
                # the residual rearrangement
                cal_end = (
                    first_fc_idx
                    + idx_ * rel_stride
                    - (math.ceil(horizon_ocs / cal_stride) - 1)
                )
                # optionally, use only `cal_length` residuals
                cal_start = cal_end - cal_length if cal_length is not None else None

                # calibrate and apply interval to the forecasts
                q_hat_ = self._calibrate_interval(res[:, :, cal_start:cal_end])
                vals = self._apply_interval(pred_vals_, q_hat_)

                # optionally, generate samples from the intervals
                if not predict_likelihood_parameters:
                    vals = sample_from_quantiles(
                        vals, self.quantiles, num_samples=num_samples
                    )
                return vals

            # historical conformal prediction
            # for each forecast, compute calibrated quantile intervals based on past residuals
            if last_points_only:
                inner_iterator = enumerate(
                    s_hfcs.all_values(copy=False)[first_fc_idx:last_fc_idx:rel_stride]
                )
            else:
                inner_iterator = enumerate(s_hfcs[first_fc_idx:last_fc_idx:rel_stride])

            comp_names_out = (
                self.likelihood.component_names(series_)
                if predict_likelihood_parameters
                else None
            )
            if len(series) == 1:
                # only use progress bar if there's no outer loop
                inner_iterator = _build_tqdm_iterator(
                    inner_iterator,
                    verbose,
                    total=(last_fc_idx - 1 - first_fc_idx) // rel_stride + 1,
                    desc="conformal forecasts",
                )

            if last_points_only:
                for idx, pred_vals in inner_iterator:
                    pred_vals = np.expand_dims(pred_vals, 0)
                    cp_pred = conformal_predict(idx, pred_vals)
                    cp_preds.append(cp_pred)
                cp_preds = _build_forecast_series(
                    points_preds=np.concatenate(cp_preds, axis=0),
                    input_series=series_,
                    custom_columns=comp_names_out,
                    time_index=generate_index(
                        start=s_hfcs._time_index[first_fc_idx],
                        length=len(cp_preds),
                        freq=series_.freq * stride,
                        name=series_._time_index.name,
                    ),
                    with_static_covs=not predict_likelihood_parameters,
                    with_hierarchy=False,
                )
            else:
                for idx, pred in inner_iterator:
                    pred_vals = pred.all_values(copy=False)
                    cp_pred = conformal_predict(idx, pred_vals)
                    cp_pred = _build_forecast_series(
                        points_preds=cp_pred,
                        input_series=series_,
                        custom_columns=comp_names_out,
                        time_index=pred._time_index,
                        with_static_covs=not predict_likelihood_parameters,
                        with_hierarchy=False,
                    )
                    cp_preds.append(cp_pred)
            cp_hfcs.append(cp_preds)
        return cp_hfcs

    def _clean(self) -> Self:
        """Cleans the model and sub-model."""
        cleaned_model = super()._clean()
        cleaned_model.model = cleaned_model.model._clean()
        return cleaned_model

    def save(
        self,
        path: Optional[Union[str, os.PathLike, BinaryIO]] = None,
        clean: bool = False,
        **pkl_kwargs,
    ) -> None:
        """
        Saves the conformal model under a given path or file handle.

        Additionally, two files are stored if `self.model` is a `TorchForecastingModel`.

        Example for saving and loading a :class:`ConformalNaiveModel`:

            .. highlight:: python
            .. code-block:: python

                from darts.datasets import AirPassengersDataset
                from darts.models import ConformalNaiveModel, LinearRegressionModel

                series = AirPassengersDataset().load()
                forecasting_model = LinearRegressionModel(lags=4).fit(series)

                model = ConformalNaiveModel(
                    model=forecasting_model,
                    quantiles=[0.1, 0.5, 0.9],
                )

                model.save("my_model.pkl")
                model_loaded = ConformalNaiveModel.load("my_model.pkl")
            ..

        Parameters
        ----------
        path
            Path or file handle under which to save the ensemble model at its current state. If no path is specified,
            the ensemble model is automatically saved under ``"{ConformalNaiveModel}_{YYYY-mm-dd_HH_MM_SS}.pkl"``.
            If the forecasting model is a `TorchForecastingModel`, two files (model object and checkpoint) are saved
            under ``"{path}.{ModelClass}.pt"`` and ``"{path}.{ModelClass}.ckpt"``.
        clean
            Whether to store a cleaned version of the model. If `True`, the training series and covariates are removed.
            If the underlying forecasting `model` is a `TorchForecastingModel`, will additionally remove all Lightning
            Trainer-related parameters.

            Note: After loading a model stored with `clean=True`, a `series` must be passed 'predict()',
            `historical_forecasts()` and other forecasting methods.
        pkl_kwargs
            Keyword arguments passed to `pickle.dump()`
        """

        if path is None:
            # default path
            path = self._default_save_path() + ".pkl"

        super().save(path, clean=clean, **pkl_kwargs)

        if TORCH_AVAILABLE and issubclass(type(self.model), TorchForecastingModel):
            path_tfm = f"{path}.{type(self.model).__name__}.pt"
            self.model.save(path=path_tfm, clean=clean)

    @staticmethod
    def load(
        path: Union[str, os.PathLike, BinaryIO],
        pl_trainer_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> "ConformalModel":
        """
        Loads a model from a given path or file handle.

        Parameters
        ----------
        path
            Path or file handle from which to load the model.
        pl_trainer_kwargs
            Only effective if the underlying forecasting model is a `TorchForecastingModel`.
            Optionally, a set of kwargs to create a new Lightning Trainer used to configure the model for downstream
            tasks (e.g. prediction).
            Some examples include specifying the batch size or moving the model to CPU/GPU(s). Check the
            `Lightning Trainer documentation <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_
            for more information about the supported kwargs.
        **kwargs
            Only effective if the underlying forecasting model is a `TorchForecastingModel`.
            Additional kwargs for PyTorch Lightning's :func:`LightningModule.load_from_checkpoint()` method,
            For more information, read the `official documentation <https://pytorch-lightning.readthedocs.io/en/stable/
            common/lightning_module.html#load-from-checkpoint>`_.
        """
        model: ConformalModel = GlobalForecastingModel.load(path)

        if TORCH_AVAILABLE and issubclass(type(model.model), TorchForecastingModel):
            path_tfm = f"{path}.{type(model.model).__name__}.pt"
            model.model = TorchForecastingModel.load(
                path_tfm,
                pl_trainer_kwargs=pl_trainer_kwargs,
                **kwargs,
            )
        return model

    @abstractmethod
    def _calibrate_interval(
        self, residuals: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Computes the lower and upper calibrated forecast intervals based on residuals.

        Parameters
        ----------
        residuals
            The residuals are expected to have shape (horizon, n components, n historical forecasts * n samples)
        """

    @abstractmethod
    def _apply_interval(self, pred: np.ndarray, q_hat: tuple[np.ndarray, np.ndarray]):
        """Applies the calibrated interval to the predicted quantiles. Returns an array with `len(quantiles)`
        conformalized quantile predictions (lower quantiles, model forecast, upper quantiles) per component.

        E.g. output is `(target1_q1, target1_pred, target1_q2, target2_q1, ...)`
        """

    @property
    @abstractmethod
    def _residuals_metric(self) -> tuple[METRIC_TYPE, Optional[dict]]:
        """Gives the "per time step" metric and optional metric kwargs used to compute residuals /
        non-conformity scores."""

    def _historical_forecasts_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        super()._historical_forecasts_sanity_checks(*args, **kwargs, is_conformal=True)

    @property
    def output_chunk_length(self) -> Optional[int]:
        # conformal models can predict any horizon if the calibration set is large enough
        return None

    @property
    def output_chunk_shift(self) -> int:
        return self.model.output_chunk_shift

    @property
    def _model_encoder_settings(self):
        raise NotImplementedError(f"not supported by `{self.__class__.__name__}`.")

    @property
    def extreme_lags(
        self,
    ) -> tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        int,
        Optional[int],
    ]:
        raise NotImplementedError(f"not supported by `{self.__class__.__name__}`.")

    @property
    def min_train_series_length(self) -> int:
        raise NotImplementedError(f"not supported by `{self.__class__.__name__}`.")

    @property
    def min_train_samples(self) -> int:
        raise NotImplementedError(f"not supported by `{self.__class__.__name__}`.")

    @property
    def supports_multivariate(self) -> bool:
        return self.model.supports_multivariate

    @property
    def supports_past_covariates(self) -> bool:
        return self.model.supports_past_covariates

    @property
    def supports_future_covariates(self) -> bool:
        return self.model.supports_future_covariates

    @property
    def supports_static_covariates(self) -> bool:
        return self.model.supports_static_covariates

    @property
    def supports_sample_weight(self) -> bool:
        return self.model.supports_sample_weight

    @property
    def supports_likelihood_parameter_prediction(self) -> bool:
        return True

    @property
    def supports_probabilistic_prediction(self) -> bool:
        return True

    @property
    def uses_past_covariates(self) -> bool:
        return self.model.uses_past_covariates

    @property
    def uses_future_covariates(self) -> bool:
        return self.model.uses_future_covariates

    @property
    def uses_static_covariates(self) -> bool:
        return self.model.uses_static_covariates

    @property
    def considers_static_covariates(self) -> bool:
        return self.model.considers_static_covariates

    @property
    def likelihood(self) -> Likelihood:
        return self._likelihood


class ConformalNaiveModel(ConformalModel):
    def __init__(
        self,
        model: GlobalForecastingModel,
        quantiles: list[float],
        symmetric: bool = True,
        cal_length: Optional[int] = None,
        cal_stride: int = 1,
        cal_num_samples: int = 500,
        random_state: Optional[int] = None,
    ):
        """Naive Conformal Prediction Model.

        A probabilistic model that adds calibrated intervals around the median forecast from a pre-trained
        global forecasting model. It does not have to be trained and can generated calibrated forecasts
        directly using the underlying trained forecasting model. It supports two symmetry modes:

        - `symmetric=True`:
            - The lower and upper interval bounds are calibrated with the same magnitude.
            - Non-conformity scores: uses metric `ae()` (see absolute error :func:`~darts.metrics.metrics.ae`) to
              compute the non-conformity scores on the calibration set.
        - `symmetric=False`
            - The lower and upper interval bounds are calibrated separately.
            - Non-conformity scores: uses metric `err()` (see error :func:`~darts.metrics.metrics.err`) to compute the
              non-conformity scores on the calibration set for the upper bounds, an `-err()` for the lower bounds.

        Since it is a probabilistic model, you can generate forecasts in two ways (when calling `predict()`,
        `historical_forecasts()`, ...):

        - Predict the calibrated quantile intervals directly: Pass parameters `predict_likelihood_parameters=True`, and
          `num_samples=1` to the forecast method.
        - Predict stochastic samples from the calibrated quantile intervals: Pass parameters
          `predict_likelihood_parameters=False`, and `num_samples>>1` to the forecast method.

        Conformal models can be applied to any of Darts' global forecasting model, as long as the model has been
        fitted before. In general the workflow of the models to produce one calibrated forecast/prediction is as
        follows:

        - Extract a calibration set: The calibration set for each conformal forecast is automatically extracted from
          the most recent past of your input series relative to the forecast start point. The number of calibration
          examples (forecast errors / non-conformity scores) to consider can be defined at model creation
          with parameter `cal_length`. Note that when using `cal_stride>1`, a longer history is required since
          the calibration examples are generated with stridden historical forecasts.
        - Generate historical forecasts on the calibration set (using the forecasting model) with a stride `cal_stride`.
        - Compute the errors/non-conformity scores (as defined above) on these historical forecasts
        - Compute the quantile values from the errors / non-conformity scores (using our desired quantiles set at model
          creation with parameter `quantiles`).
        - Compute the conformal prediction: Using these quantile values, add calibrated intervals to the forecasting
          model's predictions.

        Some notes:

        - When computing `historical_forecasts()`, `backtest()`, `residuals()`, ... the above is applied for each
          forecast (the forecasting model's historical forecasts are only generated once for efficiency).
        - For multi-horizon forecasts, the above is applied for each step in the horizon separately.

        Parameters
        ----------
        model
            A pre-trained global forecasting model. See the list of models
            `here <https://unit8co.github.io/darts/#forecasting-models>`_.
        quantiles
            A list of quantiles centered around the median `q=0.5` to use. For example quantiles
            [0.1, 0.2, 0.5, 0.8 0.9] correspond to two intervals with (0.9 - 0.1) = 80%, and (0.8 - 0.2) 60% coverage
            around the median (model forecast).
        symmetric
            Whether to use symmetric non-conformity scores. If `True`, uses metric `ae()` (see
            :func:`~darts.metrics.metrics.ae`) to compute the non-conformity scores. If `False`, uses metric `-err()`
            (see :func:`~darts.metrics.metrics.err`) for the lower, and `err()` for the upper quantile interval bound.
        cal_length
            The number of past forecast errors / non-conformity scores to use as calibration for each conformal
            forecast (and each step in the horizon). If `None`, considers all scores.
        cal_stride
            The stride to apply when computing the historical forecasts and non-conformity scores on the calibration
            set. The actual conformal forecasts can have a different stride given with parameter `stride` in downstream
            tasks (e.g. historical forecasts, backtest, ...)
        cal_num_samples
            The number of samples to generate for each calibration forecast (if `model` is a probabilistic forecasting
            model). The non-conformity scores are computed on the quantile values of these forecasts (using quantiles
            `quantiles`). Uses `1` for deterministic models. The actual conformal forecasts can have a different number
            of samples given with parameter `num_samples` in downstream tasks (e.g. predict, historical forecasts, ...).
        random_state
            Control the randomness of probabilistic conformal forecasts (sample generation) across different runs.
        """
        super().__init__(
            model=model,
            quantiles=quantiles,
            symmetric=symmetric,
            cal_length=cal_length,
            cal_num_samples=cal_num_samples,
            random_state=random_state,
            cal_stride=cal_stride,
        )

    def _calibrate_interval(
        self, residuals: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        def q_hat_from_residuals(residuals_):
            # compute quantiles of shape (forecast horizon, n components, n quantile intervals)
            return np.quantile(
                residuals_,
                q=self.interval_range_sym,
                method="higher",
                axis=2,
            ).transpose((1, 2, 0))

        # residuals shape (horizon, n components, n past forecasts)
        if self.symmetric:
            # symmetric (from metric `ae()`)
            q_hat = q_hat_from_residuals(residuals)
            return -q_hat, q_hat[:, :, ::-1]
        else:
            # asymmetric (from metric `err()`)
            q_hat = q_hat_from_residuals(
                np.concatenate([-residuals, residuals], axis=1)
            )
            n_comps = residuals.shape[1]
            return -q_hat[:, :n_comps, :], q_hat[:, n_comps:, ::-1]

    def _apply_interval(self, pred: np.ndarray, q_hat: tuple[np.ndarray, np.ndarray]):
        # convert stochastic predictions to median
        if pred.shape[2] != 1:
            pred = np.expand_dims(np.quantile(pred, 0.5, axis=2), -1)
        # shape (forecast horizon, n components, n quantiles)
        pred = np.concatenate([pred + q_hat[0], pred, pred + q_hat[1]], axis=2)
        # -> (forecast horizon, n components * n quantiles)
        return pred.reshape(len(pred), -1)

    @property
    def _residuals_metric(self) -> tuple[METRIC_TYPE, Optional[dict]]:
        return (metrics.ae if self.symmetric else metrics.err), None


class ConformalQRModel(ConformalModel):
    def __init__(
        self,
        model: GlobalForecastingModel,
        quantiles: list[float],
        symmetric: bool = True,
        cal_length: Optional[int] = None,
        cal_stride: int = 1,
        cal_num_samples: int = 500,
        random_state: Optional[int] = None,
    ):
        """Conformalized Quantile Regression Model.

        A probabilistic model that calibrates the quantile predictions from a pre-trained probabilistic global
        forecasting model. It does not have to be trained and can generated calibrated forecasts
        directly using the underlying trained forecasting model. It supports two symmetry modes:

        - `symmetric=True`:
            - The lower and upper quantile predictions are calibrated with the same magnitude.
            - Non-conformity scores: uses metric `incs_qr(symmetric=True)` (see Non-Conformity Score for Quantile
              Regression :func:`~darts.metrics.metrics.incs_qr`) to compute the non-conformity scores on the calibration
              set.
        - `symmetric=False`
            - The lower and upper quantile predictions are calibrated separately.
            - Non-conformity scores: uses metric `incs_qr(symmetric=False)` (see Non-Conformity Score for Quantile
              Regression :func:`~darts.metrics.metrics.incs_qr`) to compute the non-conformity scores for the upper and
              lower bound separately.

        Since it is a probabilistic model, you can generate forecasts in two ways (when calling `predict()`,
        `historical_forecasts()`, ...):

        - Predict the calibrated quantile intervals directly: Pass parameters `predict_likelihood_parameters=True`, and
          `num_samples=1` to the forecast method.
        - Predict stochastic samples from the calibrated quantile intervals: Pass parameters
          `predict_likelihood_parameters=False`, and `num_samples>>1` to the forecast method.

        Conformal models can be applied to any of Darts' global forecasting model, as long as the model has been
        fitted before. In general the workflow of the models to produce one calibrated forecast/prediction is as
        follows:

        - Extract a calibration set: The calibration set for each conformal forecast is automatically extracted from
          the most recent past of your input series relative to the forecast start point. The number of calibration
          examples (forecast errors / non-conformity scores) to consider can be defined at model creation with
          parameter `cal_length`. Note that when using `cal_stride>1`, a longer history is required since the
          calibration examples are generated with stridden historical forecasts.
        - Generate historical forecasts (quantile predictions) on the calibration set (using the forecasting model)
          with a stride `cal_stride`.
        - Compute the errors/non-conformity scores (as defined above) on these historical quantile predictions
        - Compute the quantile values from the errors / non-conformity scores (using our desired quantiles set at model
          creation with parameter `quantiles`).
        - Compute the conformal prediction: Using these quantile values, calibrate the predicted quantiles from the
          forecasting model's predictions.

        Some notes:

        - When computing `historical_forecasts()`, `backtest()`, `residuals()`, ... the above is applied for each
          forecast (the forecasting model's historical forecasts are only generated once for efficiency).
        - For multi-horizon forecasts, the above is applied for each step in the horizon separately.

        Parameters
        ----------
        model
            A pre-trained global forecasting model. See the list of models
            `here <https://unit8co.github.io/darts/#forecasting-models>`_.
        quantiles
            A list of quantiles centered around the median `q=0.5` to use. For example quantiles
            [0.1, 0.2, 0.5, 0.8 0.9] correspond to two intervals with (0.9 - 0.1) = 80%, and (0.8 - 0.2) 60% coverage
            around the median (model forecast).
        symmetric
            Whether to use symmetric non-conformity scores. If `True`, uses symmetric metric
            `incs_qr(..., symmetric=True)` (see :func:`~darts.metrics.metrics.incs_qr`) to compute the non-conformity
            scores. If `False`, uses asymmetric metric `incs_qr(..., symmetric=False)` with individual scores for the
            lower- and upper quantile interval bounds.
        cal_length
            The number of past forecast errors / non-conformity scores to use as calibration for each conformal
            forecast (and each step in the horizon). If `None`, considers all scores.
        cal_stride
            The stride to apply when computing the historical forecasts and non-conformity scores on the calibration
            set. The actual conformal forecasts can have a different stride given with parameter `stride` in downstream
            tasks (e.g. historical forecasts, backtest, ...)
        cal_num_samples
            The number of samples to generate for each calibration forecast (if `model` is a probabilistic forecasting
            model). The non-conformity scores are computed on the quantile values of these forecasts (using quantiles
            `quantiles`). Uses `1` for deterministic models. The actual conformal forecasts can have a different number
            of samples given with parameter `num_samples` in downstream tasks (e.g. predict, historical forecasts, ...).
        random_state
            Control the randomness of probabilistic conformal forecasts (sample generation) across different runs.
        """
        if not model.supports_probabilistic_prediction:
            raise_log(
                ValueError(
                    "`model` must support probabilistic forecasting. Consider using a `likelihood` at "
                    "forecasting model creation, or use another conformal model."
                ),
                logger=logger,
            )
        super().__init__(
            model=model,
            quantiles=quantiles,
            symmetric=symmetric,
            cal_length=cal_length,
            cal_num_samples=cal_num_samples,
            random_state=random_state,
            cal_stride=cal_stride,
        )

    def _calibrate_interval(
        self, residuals: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        n_comps = residuals.shape[1] // (
            len(self.interval_range) * (1 + int(not self.symmetric))
        )
        n_intervals = len(self.interval_range)

        def q_hat_from_residuals(residuals_):
            # TODO: is there a more efficient way?
            # compute quantiles with shape (horizon, n components, n quantile intervals)
            # over all past residuals
            q_hat_tmp = np.quantile(
                residuals_, q=self.interval_range_sym, method="higher", axis=2
            ).transpose((1, 2, 0))
            q_hat_ = np.empty((len(residuals_), n_comps, n_intervals))
            for i in range(n_intervals):
                for c in range(n_comps):
                    q_hat_[:, c, i] = q_hat_tmp[:, i + c * n_intervals, i]
            return q_hat_

        if self.symmetric:
            # symmetric has one nc-score per interval (from metric `incs_qr(symmetric=True)`)
            # residuals shape (horizon, n components * n intervals, n past forecasts)
            q_hat = q_hat_from_residuals(residuals)
            return -q_hat, q_hat[:, :, ::-1]
        else:
            # asymmetric has two nc-score per interval (for lower and upper quantiles, from metric
            # `incs_qr(symmetric=False)`)
            # lower and upper residuals are concatenated along axis=1;
            # residuals shape (horizon, n components * n intervals * 2, n past forecasts)
            half_idx = residuals.shape[1] // 2
            q_hat_lo = q_hat_from_residuals(residuals[:, :half_idx])
            q_hat_hi = q_hat_from_residuals(residuals[:, half_idx:])
            return -q_hat_lo, q_hat_hi[:, :, ::-1]

    def _apply_interval(self, pred: np.ndarray, q_hat: tuple[np.ndarray, np.ndarray]):
        # get quantile predictions with shape (n times, n components, n quantiles)
        pred = np.quantile(pred, self.quantiles, axis=2).transpose((1, 2, 0))
        # shape (forecast horizon, n components, n quantiles)
        pred = np.concatenate(
            [
                pred[:, :, : self.idx_median] + q_hat[0],  # lower quantiles
                pred[:, :, self.idx_median : self.idx_median + 1],  # model forecast
                pred[:, :, self.idx_median + 1 :] + q_hat[1],  # upper quantiles
            ],
            axis=2,
        )
        # -> (forecast horizon, n components * n quantiles)
        return pred.reshape(len(pred), -1)

    @property
    def _residuals_metric(self) -> tuple[METRIC_TYPE, Optional[dict]]:
        return metrics.incs_qr, {
            "q_interval": self.q_interval,
            "symmetric": self.symmetric,
        }


def _get_calibration_hfc_start(
    series: Sequence[TimeSeries],
    horizon: int,
    output_chunk_shift: int,
    cal_length: Optional[int],
    cal_stride: int,
    start: Optional[Union[pd.Timestamp, int, Literal["end"]]],
    start_format: Literal["position", "value"],
) -> tuple[Optional[Union[int, pd.Timestamp]], Literal["position", "value"]]:
    """Find the calibration start point (CSP) (for historical forecasts on calibration set).

    - If `start=None`, the CSP is also `None` (all possible hfcs).
    - If `start="end"` (when calling `predict()`), returns the CSP as a positional index relative to the end of the
      series (<0).
    - Otherwise (when calling `historical_forecasts()`), the CSP is the start value (`start_format="value"`) or start
      position (`start_format="position"`) adjusted by the positions computed for the case above.

    If this function is called from `historical_forecasts`, the sanity checks guarantee the following:

    - `start` cannot be a `float`
    - when `start_format='value'`, all `series` have the same frequency
    """
    if start is None:
        return start, start_format

    cal_start_format: Literal["position", "value"]
    horizon_ocs = horizon + output_chunk_shift
    if cal_length is not None:
        # we only need `cal_length` forecasts with stride `cal_stride` before the `predict()` start point;
        # the last valid calibration forecast must start at least `horizon_ocs` before `predict()` start
        add_steps = math.ceil(horizon_ocs / cal_stride) - 1
        start_idx_rel = -cal_stride * (cal_length + add_steps)
        cal_start_format = "position"
    elif cal_stride > 1:
        # we need all forecasts with stride `cal_stride` before the `predict()` start point
        max_len_series = max(len(series_) for series_ in series)
        start_idx_rel = -cal_stride * math.ceil(max_len_series / cal_stride)
        cal_start_format = "position"
    else:
        # we need all possible forecasts with `cal_stride=1`
        start_idx_rel, cal_start_format = None, "value"

    if start == "end":
        # `predict()` is relative to the end
        return start_idx_rel, cal_start_format

    # `historical_forecasts()` is relative to `start`
    start_is_position = isinstance(start, (int, np.int64)) and (
        start_format == "position" or series[0]._has_datetime_index
    )
    cal_start_format = start_format
    if start_idx_rel is None:
        cal_start = start_idx_rel
    elif start_is_position:
        cal_start = start + start_idx_rel
        # if start switches sign, it would be relative to the end;
        # correct it to be positive (relative to beginning)
        if cal_start < 0 <= start:
            cal_start += math.ceil(abs(cal_start) / cal_stride) * cal_stride
    else:
        cal_start = start + start_idx_rel * series[0].freq
    return cal_start, cal_start_format



"""
NLL Cauchy Scorer
-----------------

Cauchy distribution negative log-likelihood Scorer.

The anomaly score is the negative log likelihood of the actual series values
under a Cauchy distribution estimated from the stochastic prediction.
"""

import numpy as np
from scipy.stats import cauchy

from darts.ad.scorers.scorers import NLLScorer


class CauchyNLLScorer(NLLScorer):
    def __init__(self, window: int = 1) -> None:
        """NLL Cauchy Scorer"""
        super().__init__(window=window)

    def __str__(self):
        return "CauchyNLLScorer"

    def _score_core_nllikelihood(
        self, vals: np.ndarray, pred_vals: np.ndarray
    ) -> np.ndarray:
        params = np.apply_along_axis(cauchy.fit, axis=1, arr=pred_vals)
        return -cauchy.logpdf(vals, loc=params[:, 0], scale=params[:, 1])



import itertools

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.tests.conftest import POLARS_AVAILABLE
from darts.tests.test_timeseries import TestTimeSeries
from darts.utils.utils import freqs

if POLARS_AVAILABLE:
    import polars as pl
else:
    pl = None


class TestTimeSeriesMultivariate:
    times1 = pd.date_range("20130101", "20130110")
    times2 = pd.date_range("20130206", "20130215")
    dataframe1 = pd.DataFrame(
        {"0": range(10), "1": range(5, 15), "2": range(10, 20)}, index=times1
    )
    dataframe2 = pd.DataFrame(
        {"0": np.arange(1, 11), "1": np.arange(1, 11) * 3, "2": np.arange(1, 11) * 5},
        index=times1,
    )
    dataframe3 = pd.DataFrame(
        {
            "0": np.arange(1, 11),
            "1": np.arange(11, 21),
        },
        index=times2,
    )
    dataframe4 = pd.DataFrame(
        {
            "0": [1, 1, np.nan, 1, 1, 1, 1, 1, 1, 1],
            "1": [1, 1, np.nan, 1, 1, np.nan, np.nan, 1, 1, 1],
            "2": [1, 1, np.nan, 1, 1, np.nan, np.nan, np.nan, np.nan, 1],
        },
        index=times2,
    )
    series1 = TimeSeries.from_dataframe(dataframe1)
    series2 = TimeSeries.from_dataframe(dataframe2)
    series3 = TimeSeries.from_dataframe(dataframe3)
    series4 = TimeSeries.from_dataframe(dataframe4)

    def test_creation(self):
        series_test = TimeSeries.from_dataframe(self.dataframe1)
        assert np.all(series_test.to_dataframe().values == self.dataframe1.values)

        # Series cannot be lower than three without passing frequency as argument to constructor
        with pytest.raises(ValueError):
            TimeSeries(self.dataframe1.iloc[:2, :])
        TimeSeries.from_dataframe(self.dataframe1.iloc[:2, :], freq="D")

    @pytest.mark.skipif(not POLARS_AVAILABLE, reason="requires polars")
    def test_polars_creation(self):
        pl_df = pl.DataFrame(
            data={
                "time": pd.date_range(start="2023-01-01", periods=10, freq="D"),
                "test_float": [float(i) for i in range(10)],
                "test_int": range(10),
            }
        )
        # with a `time_col` no warning is raised
        ts = TimeSeries.from_dataframe(pl_df, time_col="time")
        ts_pl_df = ts.to_dataframe(backend="polars", time_as_index=False)
        assert ts_pl_df.equals(pl_df)

        # darts converts everything to float (test_int)
        assert ts_pl_df.dtypes != pl_df.dtypes
        dtypes_expected = pl_df.dtypes[:2] + [pl_df.dtypes[1]]
        assert ts_pl_df.dtypes == dtypes_expected

    def test_eq(self):
        seriesA = TimeSeries.from_dataframe(self.dataframe1)
        assert self.series1 == seriesA
        assert not (self.series1 != seriesA)

        # with different dates
        dataframeB = self.dataframe1.copy()
        dataframeB.index = pd.date_range("20130102", "20130111")
        seriesB = TimeSeries.from_dataframe(dataframeB)
        assert not (self.series1 == seriesB)

        # with one different value
        dataframeC = self.dataframe1.copy()
        dataframeC.iloc[2, 2] = 0
        seriesC = TimeSeries.from_dataframe(dataframeC)
        assert not (self.series1 == seriesC)

    def test_rescale(self):
        with pytest.raises(ValueError):
            self.series1.rescale_with_value(1)

        seriesA = self.series2.rescale_with_value(0)
        assert np.all(seriesA.values() == 0).all()

        seriesB = self.series2.rescale_with_value(1)
        assert seriesB == TimeSeries.from_dataframe(
            pd.DataFrame(
                {
                    "0": np.arange(1, 11),
                    "1": np.arange(1, 11),
                    "2": np.arange(1, 11),
                },
                index=self.dataframe2.index,
            ).astype(float)
        )

    def test_slice(self):
        TestTimeSeries.helper_test_slice(self, self.series1)

    def test_split(self):
        TestTimeSeries.helper_test_split(self, self.series1)

    def test_drop(self):
        TestTimeSeries.helper_test_drop(self, self.series1)

    @pytest.mark.parametrize(
        "config", itertools.product(["D", "2D", 1, 2], [False, True])
    )
    def test_intersect(self, config):
        freq, mixed_freq = config
        TestTimeSeries.helper_test_intersect(freq, mixed_freq, is_univariate=False)

    def test_shift(self):
        TestTimeSeries.helper_test_shift(self, self.series1)

    def test_append(self):
        TestTimeSeries.helper_test_append(self, self.series1)

    def test_append_values(self):
        TestTimeSeries.helper_test_append_values(self, self.series1)

    def test_prepend(self):
        TestTimeSeries.helper_test_prepend(self, self.series1)

    def test_prepend_values(self):
        TestTimeSeries.helper_test_prepend_values(self, self.series1)

    def test_strip(self):
        dataframe1 = pd.DataFrame(
            {
                "0": 2 * [np.nan] + list(range(7)) + [np.nan],
                "1": [np.nan] + list(range(7)) + 2 * [np.nan],
            },
            index=self.times1,
        )
        series1 = TimeSeries.from_dataframe(dataframe1)

        assert (series1.strip().time_index == self.times1[1:-1]).all()
        assert (series1.strip(how="any").time_index == self.times1[2:-2]).all()

    """
    Testing new multivariate methods.
    """

    def test_stack(self):
        with pytest.raises(ValueError):
            self.series1.stack(self.series3)
        seriesA = self.series1.stack(self.series2)
        dataframeA = pd.concat([self.dataframe1, self.dataframe2], axis=1)
        dataframeA.columns = [
            "0",
            "1",
            "2",
            "0_1",
            "1_1",
            "2_1",
        ]  # the names to expect after stacking
        assert (seriesA.to_dataframe() == dataframeA).all().all()
        assert seriesA.values().shape == (
            len(self.dataframe1),
            len(self.dataframe1.columns) + len(self.dataframe2.columns),
        )

    def test_univariate_component(self):
        with pytest.raises(IndexError):
            self.series1.univariate_component(-5)
        with pytest.raises(IndexError):
            self.series1.univariate_component(3)
        seriesA = self.series1.univariate_component(1)
        assert seriesA == TimeSeries.from_times_and_values(
            self.times1, range(5, 15), columns=["1"]
        )
        seriesB = (
            self.series1.univariate_component(0)
            .stack(seriesA)
            .stack(self.series1.univariate_component(2))
        )
        assert self.series1 == seriesB

    def test_add_datetime_attribute(self):
        """datetime_attributes are 0-indexed (shift is applied when necessary)"""
        seriesA = self.series1.add_datetime_attribute("day")
        assert seriesA.width == self.series1.width + 1
        assert set(
            seriesA.to_dataframe().iloc[:, seriesA.width - 1].values.flatten()
        ) == set(range(0, 10))
        seriesB = self.series3.add_datetime_attribute("day", True)
        assert seriesB.width == self.series3.width + 31
        assert set(
            seriesB.to_dataframe().iloc[:, self.series3.width :].values.flatten()
        ) == {0, 1}
        seriesC = self.series1.add_datetime_attribute("month", True)
        assert seriesC.width == self.series1.width + 12
        seriesD = TimeSeries.from_times_and_values(
            pd.date_range("20130206", "20130430"), range(84)
        )
        seriesD = seriesD.add_datetime_attribute("month", True)
        assert seriesD.width == 13
        assert sum(seriesD.values().flatten()) == sum(range(84)) + 84
        assert sum(seriesD.values()[:, 1 + 3]) == 30
        assert sum(seriesD.values()[:, 1 + 1]) == 23

        # test cyclic
        times_month = pd.date_range("20130101", "20140610")

        seriesE = TimeSeries.from_times_and_values(
            times_month, np.repeat(0.1, len(times_month))
        )
        seriesF = seriesE.add_datetime_attribute("day", cyclic=True)

        values_sin = seriesF.values()[:, 1]
        values_cos = seriesF.values()[:, 2]

        assert np.allclose(np.add(np.square(values_sin), np.square(values_cos)), 1)

        df = seriesF.to_dataframe()
        # first day is equivalent to t=0
        df = df[df.index.day == 1]
        assert np.allclose(df["day_sin"].values, 0, atol=0.03)
        assert np.allclose(df["day_cos"].values, 1, atol=0.03)

        # second day is equivalent to t=1
        df = df[df.index.day == 2]
        assert np.allclose(df["day_sin"].values, 0.2, atol=0.03)
        assert np.allclose(df["day_cos"].values, 0.97, atol=0.03)

    def test_add_holidays(self):
        times = pd.date_range(start=pd.Timestamp("20201201"), periods=30, freq="D")
        seriesA = TimeSeries.from_times_and_values(times, range(len(times)))

        # testing for christmas and non-holiday in US
        seriesA = seriesA.add_holidays("US")
        last_column = seriesA.to_dataframe().iloc[:, seriesA.width - 1]
        assert last_column.at[pd.Timestamp("20201225")] == 1
        assert last_column.at[pd.Timestamp("20201210")] == 0
        assert last_column.at[pd.Timestamp("20201226")] == 0

        # testing for christmas and non-holiday in PL
        seriesA = seriesA.add_holidays("PL")
        last_column = seriesA.to_dataframe().iloc[:, seriesA.width - 1]
        assert last_column.at[pd.Timestamp("20201225")] == 1
        assert last_column.at[pd.Timestamp("20201210")] == 0
        assert last_column.at[pd.Timestamp("20201226")] == 1
        assert seriesA.width == 3

        # testing hourly time series
        times = pd.date_range(
            start=pd.Timestamp("20201224"), periods=50, freq=freqs["h"]
        )
        seriesB = TimeSeries.from_times_and_values(times, range(len(times)))
        seriesB = seriesB.add_holidays("US")
        last_column = seriesB.to_dataframe().iloc[:, seriesB.width - 1]
        assert last_column.at[pd.Timestamp("2020-12-25 01:00:00")] == 1
        assert last_column.at[pd.Timestamp("2020-12-24 23:00:00")] == 0

    def test_assert_univariate(self):
        with pytest.raises(AssertionError):
            self.series1._assert_univariate()
        self.series1.univariate_component(0)._assert_univariate()

    def test_first_last_values(self):
        assert self.series1.first_values().tolist() == [0, 5, 10]
        assert self.series3.last_values().tolist() == [10, 20]
        assert self.series1.univariate_component(1).first_values().tolist() == [5]
        assert self.series3.univariate_component(1).last_values().tolist() == [20]

    def test_drop_column(self):
        # testing dropping a single column
        seriesA = self.series1.drop_columns("0")
        assert "0" not in seriesA.columns.values
        assert seriesA.columns.tolist() == ["1", "2"]
        assert len(seriesA.columns) == 2

        # testing dropping multiple columns
        seriesB = self.series1.drop_columns(["0", "1"])
        assert "2" in seriesB.columns.values
        assert len(seriesB.columns) == 1

    def test_gaps(self):
        gaps1_all = self.series1.gaps(mode="all")
        assert gaps1_all.empty
        gaps1_any = self.series1.gaps(mode="any")
        assert gaps1_any.empty

        gaps4_all = self.series4.gaps(mode="all")
        assert (
            gaps4_all["gap_start"] == pd.DatetimeIndex([pd.Timestamp("20130208")])
        ).all()
        assert (
            gaps4_all["gap_end"] == pd.DatetimeIndex([pd.Timestamp("20130208")])
        ).all()
        assert gaps4_all["gap_size"].values.tolist() == [1]

        gaps4_any = self.series4.gaps(mode="any")
        assert (
            gaps4_any["gap_start"]
            == pd.DatetimeIndex([pd.Timestamp("20130208"), pd.Timestamp("20130211")])
        ).all()
        assert (
            gaps4_any["gap_end"]
            == pd.DatetimeIndex([pd.Timestamp("20130208"), pd.Timestamp("20130214")])
        ).all()
        assert gaps4_any["gap_size"].values.tolist() == [1, 4]



"""
Wasserstein Scorer
-----

Wasserstein Scorer (distance function defined between probability distributions) [1]_.
The implementations is wrapped around `scipy.stats
<https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wasserstein_distance.html>`_.

References
----------
.. [1] https://en.wikipedia.org/wiki/Wasserstein_metric
"""

from collections.abc import Sequence

import numpy as np
from scipy.stats import wasserstein_distance

from darts import metrics
from darts.ad.scorers.scorers import WindowedAnomalyScorer
from darts.logging import get_logger
from darts.metrics.metrics import METRIC_TYPE
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class WassersteinScorer(WindowedAnomalyScorer):
    def __init__(
        self,
        window: int = 10,
        component_wise: bool = False,
        window_agg: bool = True,
        diff_fn: METRIC_TYPE = metrics.ae,
    ) -> None:
        """Wasserstein Scorer

        When calling `fit(series)`, a moving window is applied, which results in a set of vectors of size `W`,
        where `W` is the window size. These vectors are kept in memory, representing the training
        distribution. The `score(series)` function will apply the same moving window.
        The Wasserstein distance is computed between the training distribution and each vector,
        resulting in an anomaly score.

        Alternatively, the scorer has the functions `fit_from_prediction()` and `score_from_prediction()`.
        Both require two series (actual and prediction), and compute a "difference" series by applying the
        function `diff_fn` (default: absolute difference). The resulting series is then passed to the
        functions `fit()` and `score()`, respectively.

        `component_wise` is a boolean parameter indicating how the model should behave with multivariate inputs
        series. If set to `True`, the model will treat each series dimension independently. If set to `False`, the model
        concatenates the dimensions in each windows of length `W` and computes a single score for all dimensions.

        **Training with** `fit()`:

        The input can be a series (univariate or multivariate) or multiple series. The series will be partitioned
        into equal size subsequences. Each subsequence has size `W * D` (features), where:

        - `W` is the size of the window given as a parameter `window`
        - `D` is the dimension of the series (`D` = 1 if univariate or if `component_wise` is set to `True`)

        For a series of length `N`, `(N - W + 1)` subsequences will be generated. The final array `X` passed to the
        underlying scorer has shape `(N - W + 1, W * D)`; or in other terms (number of samples, number of features).
        If a list of series is given of length L, each series `i` is partitioned, and all `X_i` are concatenated along
        the sample axis.

        The arrays will be kept in memory, representing the training data distribution.
        In practice, the series or list of series can for instance represent residuals than can be
        considered independent and identically distributed (iid).

        If `component_wise` is set to `True`, the algorithm will be applied to each dimension independently. For each
        dimension, a PyOD model will be trained.

        **Computing score with** `score()`:

        The input can be a series (univariate or multivariate) or a sequence of series. The given series must have the
        same dimension `D` as the data used to train the PyOD model.

        For each series, if the series is multivariate of dimension `D`:

        - if `component_wise` is set to `False`: it returns a univariate series (dimension=1). It represents
          the anomaly score of the entire series in the considered window at each timestamp.
        - if `component_wise` is set to `True`: it returns a multivariate series of dimension `D`. Each dimension
          represents the anomaly score of the corresponding component of the input.

        If the series is univariate, it returns a univariate series regardless of the parameter
        `component_wise`.

        A window of size `W` is rolled on the series with a stride equal to 1. It is the same size window `W` used
        during the training phase.
        Each value in the score series thus represents how anomalous the sample of the `W` previous values is.

        Parameters
        ----------
        window
            Size of the sliding window that represents the number of samples in the testing distribution to compare
            with the training distribution in the Wasserstein function
        component_wise
            Boolean value indicating if the score needs to be computed for each component independently (`True`)
            or by concatenating the component in the considered window to compute one score (`False`).
            Default: `False`.
        window_agg
            Boolean indicating whether the anomaly score for each time step is computed by
            averaging the anomaly scores for all windows this point is included in.
            If `False`, the anomaly score for each point is the anomaly score of its trailing window.
            Default: `True`.
        diff_fn
            The differencing function to use to transform the predicted and actual series into one series.
            The scorer is then applied to this series. Must be one of Darts per-time-step metrics (e.g.,
            :func:`~darts.metrics.metrics.ae` for the absolute difference, :func:`~darts.metrics.metrics.err` for the
            difference, :func:`~darts.metrics.metrics.se` for the squared difference, ...).
            By default, uses the absolute difference (:func:`~darts.metrics.metrics.ae`).
        """

        # TODO:
        #     - understand better the math behind the Wasserstein distance when the test distribution contains
        #     only one sample
        #     - check if there is an equivalent Wasserstein distance for d-D distributions (currently only accepts 1D)

        if type(window) is int:  # noqa: E721
            if window > 0 and window < 10:
                logger.warning(
                    f"The `window` parameter WassersteinScorer is smaller than 10 (w={window})."
                    + " The value represents the window length rolled on the series given as"
                    + " input in the `score` function. At each position, the w values will"
                    + " constitute a subset, and the Wasserstein distance between the subset"
                    + " and the train distribution will be computed. To better represent the"
                    + " constituted test distribution, the window parameter should be larger"
                    + " than 10."
                )
        super().__init__(
            is_univariate=(not component_wise),
            window=window,
            window_agg=window_agg,
            diff_fn=diff_fn,
        )

    def __str__(self):
        return "WassersteinScorer"

    def _fit_core(self, series: Sequence[TimeSeries], *args, **kwargs):
        """The training values are considered as the scorer model"""
        self.model = np.concatenate([s.all_values(copy=False) for s in series]).squeeze(
            -1
        )

        if self.is_univariate or series[0].width == 1:
            self.model = self.model.flatten()

    def _model_score_method(self, model, data: np.ndarray) -> np.ndarray:
        """Wrapper around model inference method"""
        return [wasserstein_distance(model, window_samples) for window_samples in data]



# fix pythonpath when working locally
import sys
from os import getcwd
from os.path import basename, dirname


def fix_pythonpath_if_working_locally():
    """Add the parent path to pythonpath if current working dir is darts/examples"""
    cwd = getcwd()
    if basename(cwd) == "examples":
        sys.path.insert(0, dirname(cwd))



"""
Norm Scorer
-----------

Norm anomaly score (of given order) [1]_.

References
----------
.. [1] https://en.wikipedia.org/wiki/Norm_(mathematics)
"""

import numpy as np

from darts.ad.scorers.scorers import AnomalyScorer


class NormScorer(AnomalyScorer):
    def __init__(self, ord=None, component_wise: bool = False) -> None:
        """Norm Scorer

        Returns the element-wise norm of a given order between two series' values.

        If `component_wise` is `False`, the norm is computed between vectors
        made of the series' components (one norm per timestamp).

        If `component_wise` is `True`, for any `ord` this effectively amounts to computing the absolute
        value of the difference.

        The scoring function expects two series.

        If the two series are multivariate of width `w`:

        - if `component_wise` is set to `False`: it returns a univariate series (width=1).
        - if `component_wise` is set to `True`: it returns a multivariate series of width `w`.

        If the two series are univariate, it returns a univariate series regardless of the parameter
        `component_wise`.

        Parameters
        ----------
        ord
            Order of the norm. Options are listed under 'Notes' at:
            <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>.
            Default: `None`
        component_wise
            Whether to compare components of the two series in isolation (`True`), or jointly (`False`).
            Default: `False`
        """
        self.ord = ord
        super().__init__(is_univariate=(not component_wise), window=1)

    def __str__(self):
        return f"Norm (ord={self.ord})"

    def _score_core_from_prediction(
        self,
        vals: np.ndarray,
        pred_vals: np.ndarray,
    ) -> np.ndarray:
        vals = self._extract_deterministic_values(vals, "series")
        pred_vals = self._extract_deterministic_values(pred_vals, "pred_series")
        diff = vals - pred_vals
        if not self.is_univariate:
            diff = np.abs(diff)
        else:
            diff = np.linalg.norm(diff, ord=self.ord, axis=1)
        return diff



"""
Forecasting Models
------------------

Baseline Models (`LocalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#local-forecasting-models-lfms>`_)
    - :class:`~darts.models.forecasting.baselines.NaiveMean`
    - :class:`~darts.models.forecasting.baselines.NaiveSeasonal`
    - :class:`~darts.models.forecasting.baselines.NaiveDrift`
    - :class:`~darts.models.forecasting.baselines.NaiveMovingAverage`
Global Baseline Models (`GlobalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#global-forecasting-models-gfms>`_)
    - :class:`~darts.models.forecasting.global_baseline_models.GlobalNaiveAggregate`
    - :class:`~darts.models.forecasting.global_baseline_models.GlobalNaiveDrift`
    - :class:`~darts.models.forecasting.global_baseline_models.GlobalNaiveSeasonal`
Statistical Models (`LocalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#local-forecasting-models-lfms>`_)
    - :class:`~darts.models.forecasting.arima.ARIMA`
    - :class:`~darts.models.forecasting.varima.VARIMA`
    - :class:`~darts.models.forecasting.sf_auto_arima.AutoARIMA`
    - :class:`~darts.models.forecasting.exponential_smoothing.ExponentialSmoothing`
    - :class:`~darts.models.forecasting.sf_model.StatsForecastModel`
    - :class:`~darts.models.forecasting.sf_auto_ets.AutoETS`
    - :class:`~darts.models.forecasting.sf_auto_ces.AutoCES`
    - :class:`~darts.models.forecasting.sf_auto_mfles.AutoMFLES`
    - :class:`~darts.models.forecasting.sf_tbats.TBATS`
    - :class:`~darts.models.forecasting.sf_auto_tbats.AutoTBATS`
    - :class:`~darts.models.forecasting.theta.Theta`
    - :class:`~darts.models.forecasting.theta.FourTheta`
    - :class:`~darts.models.forecasting.sf_auto_theta.AutoTheta`
    - :class:`~darts.models.forecasting.prophet_model.Prophet`
    - :class:`~darts.models.forecasting.fft.FFT`
    - :class:`~darts.models.forecasting.kalman_forecaster.KalmanForecaster`
    - :class:`~darts.models.forecasting.sf_croston.Croston`
Regression Models (`GlobalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#global-forecasting-models-gfms>`_)
    - :class:`~darts.models.forecasting.regression_model.RegressionModel`
    - :class:`~darts.models.forecasting.linear_regression_model.LinearRegressionModel`
    - :class:`~darts.models.forecasting.random_forest.RandomForest`
    - :class:`~darts.models.forecasting.lgbm.LightGBMModel`
    - :class:`~darts.models.forecasting.xgboost.XGBModel`
    - :class:`~darts.models.forecasting.catboost_model.CatBoostModel`
PyTorch (Lightning)-based Models (`GlobalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#global-forecasting-models-gfms>`_)
    - :class:`~darts.models.forecasting.rnn_model.RNNModel`
    - :class:`~darts.models.forecasting.block_rnn_model.BlockRNNModel`
    - :class:`~darts.models.forecasting.nbeats.NBEATSModel`
    - :class:`~darts.models.forecasting.nhits.NHiTSModel`
    - :class:`~darts.models.forecasting.tcn_model.TCNModel`
    - :class:`~darts.models.forecasting.transformer_model.TransformerModel`
    - :class:`~darts.models.forecasting.tft_model.TFTModel`
    - :class:`~darts.models.forecasting.dlinear.DLinearModel`
    - :class:`~darts.models.forecasting.nlinear.NLinearModel`
    - :class:`~darts.models.forecasting.tide_model.TiDEModel`
    - :class:`~darts.models.forecasting.tsmixer_model.TSMixerModel`
Ensemble Models (`GlobalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#global-forecasting-models-gfms>`_)
    - :class:`~darts.models.forecasting.baselines.NaiveEnsembleModel`
    - :class:`~darts.models.forecasting.regression_ensemble_model.RegressionEnsembleModel`
Conformal Models  (`GlobalForecastingModel <https://unit8co.github.io/darts/userguide/covariates.html#global-forecasting-models-gfms>`_)
    - :class:`~darts.models.forecasting.conformal_models.ConformalNaiveModel`
    - :class:`~darts.models.forecasting.conformal_models.ConformalQRModel`
"""



from collections.abc import Sequence
from typing import Literal, Optional, Union

import numpy as np
import pandas as pd
from numpy.lib.stride_tricks import sliding_window_view

from darts.logging import get_logger
from darts.timeseries import TimeSeries
from darts.utils import _build_tqdm_iterator
from darts.utils.data.tabularization import create_lagged_prediction_data
from darts.utils.historical_forecasts.utils import (
    _get_historical_forecast_boundaries,
)
from darts.utils.utils import generate_index

logger = get_logger(__name__)


def _optimized_historical_forecasts_last_points_only(
    model,
    series: Sequence[TimeSeries],
    past_covariates: Optional[Sequence[TimeSeries]] = None,
    future_covariates: Optional[Sequence[TimeSeries]] = None,
    num_samples: int = 1,
    start: Optional[Union[pd.Timestamp, float, int]] = None,
    start_format: Literal["position", "value"] = "value",
    forecast_horizon: int = 1,
    stride: int = 1,
    overlap_end: bool = False,
    show_warnings: bool = True,
    verbose: bool = False,
    predict_likelihood_parameters: bool = False,
    **kwargs,
) -> Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]:
    """
    Optimized historical forecasts for RegressionModel with last_points_only = True

    Rely on _check_optimizable_historical_forecasts() to check that the assumptions are verified.

    The data_transformers are applied in historical_forecasts (input and predictions)
    """
    forecasts_list = []
    iterator = _build_tqdm_iterator(
        series, verbose, total=len(series), desc="historical forecasts"
    )
    for idx, series_ in enumerate(iterator):
        past_covariates_ = past_covariates[idx] if past_covariates is not None else None
        future_covariates_ = (
            future_covariates[idx] if future_covariates is not None else None
        )
        freq = series_.freq
        forecast_components = (
            model.likelihood.component_names(series_)
            if predict_likelihood_parameters
            else series_.columns
        )

        # obtain forecastable indexes boundaries, adjust target & covariates boundaries accordingly
        (
            hist_fct_start,
            hist_fct_end,
            hist_fct_tgt_start,
            hist_fct_tgt_end,
            hist_fct_pc_start,
            hist_fct_pc_end,
            hist_fct_fc_start,
            hist_fct_fc_end,
        ) = _get_historical_forecast_boundaries(
            model=model,
            series=series_,
            series_idx=idx,
            past_covariates=past_covariates_,
            future_covariates=future_covariates_,
            start=start,
            start_format=start_format,
            forecast_horizon=forecast_horizon,
            overlap_end=overlap_end,
            stride=stride,
            freq=freq,
            show_warnings=show_warnings,
        )

        # Additional shift, to account for the model output_chunk_length
        if model.output_chunk_length != forecast_horizon and not model.multi_models:
            # used to convert the shift into the appropriate unit
            unit = freq if series_.has_datetime_index else 1

            shift = model.output_chunk_length - forecast_horizon

            hist_fct_tgt_start -= shift * unit
            hist_fct_pc_start -= (
                shift * unit if hist_fct_pc_start is not None else hist_fct_pc_start
            )
            hist_fct_fc_start -= (
                shift * unit if hist_fct_fc_start is not None else hist_fct_fc_start
            )

            hist_fct_tgt_end -= shift * unit
            hist_fct_pc_end -= (
                shift * unit if hist_fct_pc_end is not None else hist_fct_pc_end
            )
            hist_fct_fc_end -= (
                shift * unit if hist_fct_fc_end is not None else hist_fct_fc_end
            )

        X, times = create_lagged_prediction_data(
            target_series=(
                None
                if model._get_lags("target") is None
                and not model.uses_static_covariates
                else series_[hist_fct_tgt_start:hist_fct_tgt_end]
            ),
            past_covariates=(
                None
                if past_covariates_ is None
                else past_covariates_[hist_fct_pc_start:hist_fct_pc_end]
            ),
            future_covariates=(
                None
                if future_covariates_ is None
                else future_covariates_[hist_fct_fc_start:hist_fct_fc_end]
            ),
            lags=model._get_lags("target"),
            lags_past_covariates=model._get_lags("past"),
            lags_future_covariates=model._get_lags("future"),
            uses_static_covariates=model.uses_static_covariates,
            last_static_covariates_shape=model._static_covariates_shape,
            max_samples_per_ts=None,
            check_inputs=True,
            use_moving_windows=True,
            concatenate=False,
        )

        # stride can be applied directly (same for input and historical forecasts)
        X = X[0][::stride, :, 0]

        # repeat rows for probabilistic forecast
        forecast = model._predict(
            x=np.repeat(X, num_samples, axis=0),
            num_samples=num_samples,
            predict_likelihood_parameters=predict_likelihood_parameters,
            **kwargs,
        )
        # forecast has shape ((forecastable_index_length-1)*num_samples, k, n_component)
        # where k = output_chunk length if multi_models, 1 otherwise
        # reshape into (forecasted indexes, output_chunk_length, n_components, n_samples)
        forecast = np.moveaxis(
            forecast.reshape(
                X.shape[0],
                num_samples,
                model.output_chunk_length if model.multi_models else 1,
                -1,
            ),
            1,
            -1,
        )

        # extract the last sub-model forecast for each component
        if model.multi_models:
            forecast = forecast[:, forecast_horizon - 1]
        else:
            forecast = forecast[:, 0]

        if (
            stride == 1
            and model.output_chunk_length == 1
            and model.output_chunk_shift == 0
        ):
            times = times[0]
        else:
            times = generate_index(
                start=hist_fct_start
                + (forecast_horizon + model.output_chunk_shift - 1) * freq,
                length=forecast.shape[0],
                freq=freq * stride,
                name=series_.time_index.name,
            )

        forecasts_list.append(
            TimeSeries.from_times_and_values(
                times=times,
                values=forecast,
                columns=forecast_components,
                static_covariates=series_.static_covariates,
                hierarchy=series_.hierarchy,
                metadata=series_.metadata,
            )
        )
    return forecasts_list


def _optimized_historical_forecasts_all_points(
    model,
    series: Sequence[TimeSeries],
    past_covariates: Optional[Sequence[TimeSeries]] = None,
    future_covariates: Optional[Sequence[TimeSeries]] = None,
    num_samples: int = 1,
    start: Optional[Union[pd.Timestamp, float, int]] = None,
    start_format: Literal["position", "value"] = "value",
    forecast_horizon: int = 1,
    stride: int = 1,
    overlap_end: bool = False,
    show_warnings: bool = True,
    verbose: bool = False,
    predict_likelihood_parameters: bool = False,
    **kwargs,
) -> Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]:
    """
    Optimized historical forecasts for RegressionModel with last_points_only = False.

    Rely on _check_optimizable_historical_forecasts() to check that the assumptions are verified.
    """
    forecasts_list = []
    iterator = _build_tqdm_iterator(
        series, verbose, total=len(series), desc="historical forecasts"
    )
    for idx, series_ in enumerate(iterator):
        past_covariates_ = past_covariates[idx] if past_covariates is not None else None
        future_covariates_ = (
            future_covariates[idx] if future_covariates is not None else None
        )
        freq = series_.freq
        forecast_components = (
            model.likelihood.component_names(series_)
            if predict_likelihood_parameters
            else series_.columns
        )

        # obtain forecastable indexes boundaries, adjust target & covariates boundaries accordingly
        (
            hist_fct_start,
            hist_fct_end,
            hist_fct_tgt_start,
            hist_fct_tgt_end,
            hist_fct_pc_start,
            hist_fct_pc_end,
            hist_fct_fc_start,
            hist_fct_fc_end,
        ) = _get_historical_forecast_boundaries(
            model=model,
            series=series_,
            series_idx=idx,
            past_covariates=past_covariates_,
            future_covariates=future_covariates_,
            start=start,
            start_format=start_format,
            forecast_horizon=forecast_horizon,
            overlap_end=overlap_end,
            stride=stride,
            freq=freq,
            show_warnings=show_warnings,
        )

        # Additional shift, to account for the model output_chunk_length
        shift_start = 0
        if model.output_chunk_length > 1:
            # used to convert the shift into the appropriate unit
            unit = freq if series_.has_datetime_index else 1

            if not model.multi_models:
                shift_start = model.output_chunk_length - 1

                hist_fct_tgt_start -= shift_start * unit
                hist_fct_pc_start -= (
                    shift_start * unit
                    if hist_fct_pc_start is not None
                    else hist_fct_pc_start
                )
                hist_fct_fc_start -= (
                    shift_start * unit
                    if hist_fct_fc_start is not None
                    else hist_fct_fc_start
                )

        X, _ = create_lagged_prediction_data(
            target_series=(
                None
                if model._get_lags("target") is None
                and not model.uses_static_covariates
                else series_[hist_fct_tgt_start:hist_fct_tgt_end]
            ),
            past_covariates=(
                None
                if past_covariates_ is None
                else past_covariates_[hist_fct_pc_start:hist_fct_pc_end]
            ),
            future_covariates=(
                None
                if future_covariates_ is None
                else future_covariates_[hist_fct_fc_start:hist_fct_fc_end]
            ),
            lags=model._get_lags("target"),
            lags_past_covariates=model._get_lags("past"),
            lags_future_covariates=model._get_lags("future"),
            uses_static_covariates=model.uses_static_covariates,
            last_static_covariates_shape=model._static_covariates_shape,
            max_samples_per_ts=None,
            check_inputs=True,
            use_moving_windows=True,
            concatenate=False,
            show_warnings=False,
        )

        # stride must be applied post-hoc to avoid missing values
        X = X[0][:, :, 0]

        # repeat rows for probabilistic forecast
        forecast = model._predict(
            x=np.repeat(X, num_samples, axis=0),
            num_samples=num_samples,
            predict_likelihood_parameters=predict_likelihood_parameters,
            **kwargs,
        )
        # forecast has shape ((forecastable_index_length-1)*num_samples, k, n_component)
        # where k = output_chunk length if multi_models, 1 otherwise
        # reshape into (forecasted indexes, output_chunk_length, n_components, n_samples)
        forecast = np.moveaxis(
            forecast.reshape(
                X.shape[0],
                num_samples,
                model.output_chunk_length if model.multi_models else 1,
                -1,
            ),
            1,
            -1,
        )

        if model.multi_models:
            forecast = forecast[::stride, :forecast_horizon]
        else:
            # entire forecast horizon is given by multiple (previous) forecasts -> apply sliding window
            forecast = sliding_window_view(
                forecast[:, 0],
                (forecast_horizon, len(forecast_components), num_samples),
            )

            # apply stride, remove the last windows, slice output_chunk_length to keep forecast_horizon values
            if forecast_horizon != model.output_chunk_length:
                forecast = forecast[
                    : -shift_start + forecast_horizon - 1 : stride,
                    0,
                    0,
                    :forecast_horizon,
                    :,
                    :,
                ]
            # apply stride
            else:
                forecast = forecast[::stride, 0, 0, :, :, :]

        # TODO: check if faster to create in the loop
        new_times = generate_index(
            start=hist_fct_start + model.output_chunk_shift * series_.freq,
            length=forecast_horizon + (forecast.shape[0] - 1) * stride,
            freq=freq,
            name=series_.time_index.name,
        )

        forecasts_ = []
        for idx_ftc, step_fct in enumerate(
            range(0, forecast.shape[0] * stride, stride)
        ):
            forecasts_.append(
                TimeSeries.from_times_and_values(
                    times=new_times[step_fct : step_fct + forecast_horizon],
                    values=forecast[idx_ftc],
                    columns=forecast_components,
                    static_covariates=series_.static_covariates,
                    hierarchy=series_.hierarchy,
                    metadata=series_.metadata,
                )
            )

        forecasts_list.append(forecasts_)
    return forecasts_list



import pytest

from darts.utils.timeseries_generation import linear_timeseries
from darts.utils.ts_utils import (
    SeriesType,
    get_series_seq_type,
    get_single_series,
    series2seq,
)


class TestTsUtils:
    def test_series_type(self):
        assert SeriesType.NONE.value == -1
        assert SeriesType.SINGLE.value == 0
        assert SeriesType.SEQ.value == 1
        assert SeriesType.SEQ_SEQ.value == 2

        # equality works with members
        assert SeriesType.NONE == SeriesType.NONE
        assert SeriesType.SINGLE == SeriesType.SINGLE
        assert SeriesType.SEQ == SeriesType.SEQ
        assert SeriesType.SEQ_SEQ == SeriesType.SEQ_SEQ

        # inequality works with members
        assert SeriesType.SINGLE != SeriesType.SEQ
        assert SeriesType.SEQ != SeriesType.SEQ_SEQ

        # equality does not work with non-members
        with pytest.raises(ValueError) as err:
            _ = SeriesType.SINGLE == 0
        assert str(err.value).startswith("`other` must be a `SeriesType` enum.")

        # single series order is < sequence of series order < sequence of sequences of series order
        assert SeriesType.NONE < SeriesType.SINGLE < SeriesType.SEQ < SeriesType.SEQ_SEQ
        assert SeriesType.SEQ_SEQ > SeriesType.SEQ > SeriesType.SINGLE > SeriesType.NONE

    def test_get_series_seq_type(self):
        ts = linear_timeseries(length=3)
        assert get_series_seq_type(None) == SeriesType.NONE
        assert get_series_seq_type(ts) == SeriesType.SINGLE
        assert get_series_seq_type([ts]) == SeriesType.SEQ
        assert get_series_seq_type([[ts]]) == SeriesType.SEQ_SEQ

        # unknown sequence type
        with pytest.raises(ValueError) as err:
            _ = get_series_seq_type([[[ts]]])
        assert str(err.value).startswith(
            "input series must be of type `TimeSeries`, `Sequence[TimeSeries]`"
        )

        # sequence with elements different from `TimeSeries`
        with pytest.raises(ValueError) as err:
            _ = get_series_seq_type([[0.0, 1.0, 2]])
        assert str(err.value).startswith(
            "input series must be of type `TimeSeries`, `Sequence[TimeSeries]`"
        )

    def test_series2seq(self):
        ts = linear_timeseries(length=3)

        # `None` to different sequence types
        assert series2seq(None, seq_type_out=SeriesType.SINGLE) is None
        assert series2seq(None, seq_type_out=SeriesType.SEQ) is None
        assert series2seq(None, seq_type_out=SeriesType.SEQ_SEQ) is None

        # `TimeSeries` to different sequence types
        assert series2seq(ts, seq_type_out=SeriesType.SINGLE) == ts
        assert series2seq(ts, seq_type_out=SeriesType.SEQ) == [ts]
        assert series2seq(ts, seq_type_out=SeriesType.SEQ_SEQ) == [[ts]]

        # Sequence[`TimeSeries`] to different sequence types
        assert series2seq([ts], seq_type_out=SeriesType.SINGLE) == ts
        assert series2seq([ts], seq_type_out=SeriesType.SEQ) == [ts]
        assert series2seq([ts], seq_type_out=SeriesType.SEQ_SEQ) == [[ts]]

        # Sequence[`TimeSeries`, `TimeSeries`] to different sequence types
        # cannot reduce dimension since there is more than one element in SEQ
        assert series2seq([ts, ts], seq_type_out=SeriesType.SINGLE) == [ts, ts]
        assert series2seq([ts, ts], seq_type_out=SeriesType.SEQ) == [ts, ts]
        assert series2seq([ts, ts], seq_type_out=SeriesType.SEQ_SEQ) == [[ts, ts]]
        assert series2seq([ts, ts], seq_type_out=SeriesType.SEQ_SEQ, nested=True) == [
            [ts],
            [ts],
        ]

        # Sequence[Sequence[`TimeSeries`]] to different sequence types
        # SEQ_SEQ represents historical forecasts (and downstream tasks) output
        # the outer sequence represents the series axis, therefore reducing to SINGLE
        # actually returns a Sequence[`TimeSeries`]
        assert series2seq([[ts]], seq_type_out=SeriesType.SINGLE) == [ts]
        assert series2seq([[ts]], seq_type_out=SeriesType.SEQ) == [[ts]]
        assert series2seq([[ts]], seq_type_out=SeriesType.SEQ_SEQ) == [[ts]]

        # Sequence[`TimeSeries`, `TimeSeries`] to different sequence types
        # cannot reduce dimension since there is more than one element in SEQ_SEQ
        assert series2seq([[ts], [ts]], seq_type_out=SeriesType.SINGLE) == [[ts], [ts]]
        assert series2seq([[ts], [ts]], seq_type_out=SeriesType.SEQ) == [[ts], [ts]]
        assert series2seq([[ts], [ts]], seq_type_out=SeriesType.SEQ_SEQ) == [[ts], [ts]]

    def test_get_single_series(self):
        ts = linear_timeseries(length=3)
        assert get_single_series(None) is None
        assert get_single_series(ts) == ts
        assert get_single_series([ts]) == ts
        assert get_single_series([ts, ts]) == ts



"""
ARIMA
-----

Models for ARIMA (Autoregressive integrated moving average) [1]_.
The implementations is wrapped around `statsmodels <https://github.com/statsmodels/statsmodels>`_.

References
----------
.. [1] https://wikipedia.org/wiki/Autoregressive_integrated_moving_average
"""

import sys
from collections.abc import Sequence
from typing import Literal, Optional, Union

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

import numpy as np
from statsmodels.tsa.arima.model import ARIMA as staARIMA

from darts.logging import get_logger
from darts.models.forecasting.forecasting_model import (
    TransferableFutureCovariatesLocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


IntOrIntSequence: TypeAlias = Union[int, Sequence[int]]


class ARIMA(TransferableFutureCovariatesLocalForecastingModel):
    def __init__(
        self,
        p: IntOrIntSequence = 12,
        d: int = 1,
        q: IntOrIntSequence = 0,
        seasonal_order: tuple[int, IntOrIntSequence, IntOrIntSequence, int] = (
            0,
            0,
            0,
            0,
        ),
        trend: Optional[Union[Literal["n", "c", "t", "ct"], list[int]]] = None,
        random_state: Optional[int] = None,
        add_encoders: Optional[dict] = None,
    ):
        """ARIMA
        ARIMA-type models extensible with exogenous variables (future covariates)
        and seasonal components.

        Parameters
        ----------
        p : int | Sequence[int]
            Order (number of time lags) of the autoregressive model (AR).
            If a sequence of integers, specifies the exact lags to include.
        d : int
            The order of differentiation; i.e., the number of times the data
            have had past values subtracted (I).
        q : int | Sequence[int]
            The size of the moving average window (MA).
            If a sequence of integers, specifies the exact lags to include in the window.
        seasonal_order: Tuple[int | Sequence[int], int, int | Sequence[int], int]
            The (P,D,Q,s) order of the seasonal component for the AR parameters (P),
            differences (D), MA parameters (Q) and periodicity (s). D and s are always integers,
            while P and Q may either be integers or sequence of positive integers
            specifying exactly which lag orders are included.
        trend: Literal['n', 'c', 't', 'ct'] | list[int], optional
            Parameter controlling the deterministic trend. Either a string or list of integers.
            If a string, can be 'n' for no trend, 'c' for a constant term, 't' for a linear trend in time,
            and 'ct' for a constant term and linear trend.
            If a list of integers, defines a polynomial according to `numpy.poly1d` [1]_. E.g., `[1,1,0,1]` would
            translate to :math:`a + bt + ct^3`.
            Trend term of lower order than `d + D` cannot be as they would be eliminated due to the differencing
            operation.
            Default is 'c' for models without integration, and 'n' for models with integration.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.models import ARIMA
        >>> from darts.utils.timeseries_generation import datetime_attribute_timeseries
        >>> series = AirPassengersDataset().load()
        >>> # optionally, use some future covariates; e.g. the value of the month encoded as a sine and cosine series
        >>> future_cov = datetime_attribute_timeseries(series, "month", cyclic=True, add_length=6)
        >>> # define ARIMA parameters
        >>> model = ARIMA(p=12, d=1, q=2)
        >>> model.fit(series, future_covariates=future_cov)
        >>> pred = model.predict(6, future_covariates=future_cov)
        >>> pred.values()
        array([[451.36489334],
               [416.88972829],
               [443.10520391],
               [481.07892911],
               [502.11286509],
               [555.50153984]])

        References
        ----------
        .. [1] https://numpy.org/doc/stable/reference/generated/numpy.poly1d.html
        """
        super().__init__(add_encoders=add_encoders)
        self.order = p, d, q
        self.seasonal_order = seasonal_order
        self.trend = trend
        self.model = None
        self._random_state = (
            random_state
            if random_state is None
            else np.random.RandomState(random_state)
        )

    @property
    def supports_multivariate(self) -> bool:
        return False

    def _fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):
        super()._fit(series, future_covariates)

        self._assert_univariate(series)

        # storing to restore the statsmodels model results object
        self.training_historic_future_covariates = future_covariates

        m = staARIMA(
            series.values(copy=False),
            exog=future_covariates.values(copy=False) if future_covariates else None,
            order=self.order,
            seasonal_order=self.seasonal_order,
            trend=self.trend,
        )
        self.model = m.fit()

        return self

    def _predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        historic_future_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        predict_likelihood_parameters: bool = False,
        verbose: bool = False,
    ) -> TimeSeries:
        if num_samples > 1 and self.trend:
            logger.warning(
                "Trends are not well supported yet for getting probabilistic forecasts with ARIMA."
                "If you run into issues, try calling fit() with num_samples=1 or removing the trend from"
                "your model."
            )

        super()._predict(
            n, series, historic_future_covariates, future_covariates, num_samples
        )

        # updating statsmodels results object state with the new ts and covariates
        if series is not None:
            self.model = self.model.apply(
                series.values(copy=False),
                exog=(
                    historic_future_covariates.values(copy=False)
                    if historic_future_covariates
                    else None
                ),
            )

        if num_samples == 1:
            forecast = self.model.forecast(
                steps=n,
                exog=(
                    future_covariates.values(copy=False) if future_covariates else None
                ),
            )
        else:
            forecast = self.model.simulate(
                nsimulations=n,
                repetitions=num_samples,
                initial_state=self.model.states.predicted[-1, :],
                random_state=self._random_state,
                anchor="end",
                exog=(
                    future_covariates.values(copy=False) if future_covariates else None
                ),
            )

        # restoring statsmodels results object state
        if series is not None:
            self.model = self.model.apply(
                self._orig_training_series.values(copy=False),
                exog=(
                    self.training_historic_future_covariates.values(copy=False)
                    if self.training_historic_future_covariates
                    else None
                ),
            )

        return self._build_forecast_series(forecast)

    @property
    def supports_probabilistic_prediction(self) -> bool:
        return True

    @property
    def min_train_series_length(self) -> int:
        return 30



"""
AutoTBATS
---------
"""

from typing import Optional

from statsforecast.models import AutoTBATS as SFAutoTBATS

from darts.models.forecasting.sf_model import StatsForecastModel


class AutoTBATS(StatsForecastModel):
    def __init__(
        self,
        *args,
        add_encoders: Optional[dict] = None,
        quantiles: Optional[list[float]] = None,
        random_state: Optional[int] = None,
        **kwargs,
    ):
        """Auto-TBATS based on the `Statsforecasts package <https://github.com/Nixtla/statsforecast>`_.

        Automatically selects the best TBATS model from all feasible combinations of the parameters `use_boxcox`,
        `use_trend`, `use_damped_trend`, and `use_arma_errors`. Selection is made using the AIC.
        Default value for `use_arma_errors` is `True` since this enables the evaluation of models with
        and without ARMA errors. We refer to the `StatsForecast documentation
        <https://nixtlaverse.nixtla.io/statsforecast/src/core/models.html#autotbats>`_ for the exhaustive documentation
        of the arguments.

        In addition to univariate deterministic forecasting, it comes with additional support:

        - **Future covariates:** Use exogenous features to potentially improve predictive accuracy.
          Darts adds support by first regressing the series against the future covariates using a
          :class:`~darts.models.forecasting.linear_regression_model.LinearRegressionModel` model and then running the
          StatsForecast model on the in-sample residuals from this original regression. This approach was inspired by
          `this post of Stephan Kolassa <https://stats.stackexchange.com/q/220885>`_.

        - **Probabilstic forecasting:** To generate probabilistic forecasts, you can set the following
          parameters when calling :meth:`~darts.models.forecasting.sf_model.StatsForecastModel.predict`:

          - Forecast quantile values directly by setting `predict_likelihood_parameters=True`.

          - Generate sampled forecasts from these quantiles by setting `num_samples >> 1`.

        - **Transferable series forecasting:** Apply the fitted model to a new input `series` at prediction time.
          Darts adds support by first fitting a copy of the model on the new series, and then using that model to
          generate the corresponding forecast.

        .. note::
            Future covariates are not supported when the input series contain missing values.

        .. note::
            The first model call might take more time than all subsequent calls as the model relies on Numba and jit
            compilation.

        Parameters
        ----------
        args
            Positional arguments for ``statsforecasts.models.AutoTBATS``.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        quantiles
            Optionally, produce quantile predictions at `quantiles` levels when performing probabilistic forecasting
            with `num_samples > 1` or `predict_likelihood_parameters=True`.
        random_state
            Control the randomness of probabilistic conformal forecasts (sample generation) across different runs.
        kwargs
            Keyword arguments for ``statsforecasts.models.AutoTBATS``.

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.models import AutoTBATS
        >>> series = AirPassengersDataset().load()
        >>> # define AutoTBATS parameters
        >>> model = AutoTBATS(season_length=12)
        >>> model.fit(series)
        >>> pred = model.predict(6)
        >>> pred.values()
        array([[450.79653684],
               [472.09265790],
               [497.76948306],
               [510.74927369],
               [520.92224557],
               [570.33881522]])
        """
        super().__init__(
            model=SFAutoTBATS(*args, **kwargs),
            quantiles=quantiles,
            add_encoders=add_encoders,
            random_state=random_state,
        )



from typing import Optional

import numpy as np

from darts import TimeSeries


def prepare_onnx_inputs(
    model,
    series: TimeSeries,
    past_covariates: Optional[TimeSeries] = None,
    future_covariates: Optional[TimeSeries] = None,
) -> tuple[np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]:
    """Helper function to slice and concatenate the input features.

    In order to remove the dependency on the `model` argument, it can be decomposed into
    the following arguments (and simplified depending on the characteristics of the model used):
      - model_icl
      - model_ocl
      - model_uses_past_covs
      - model_uses_future_covs
      - model_uses_static_covs
    """
    past_feats, future_feats, static_feats = None, None, None
    # get input & output windows
    past_start = series.end_time() - (model.input_chunk_length - 1) * series.freq
    past_end = series.end_time()
    future_start = past_end + 1 * series.freq
    future_end = past_end + model.output_chunk_length * series.freq
    # extract all historic and future features from target, past and future covariates
    past_feats = series[past_start:past_end].values()
    if past_covariates and model.uses_past_covariates:
        # extract past covariates
        past_feats = np.concatenate(
            [past_feats, past_covariates[past_start:past_end].values()], axis=1
        )
    if future_covariates and model.uses_future_covariates:
        # extract past part of future covariates
        past_feats = np.concatenate(
            [past_feats, future_covariates[past_start:past_end].values()], axis=1
        )
        # extract future part of future covariates
        future_feats = future_covariates[future_start:future_end].values()
    # add batch dimension -> (batch, n time steps, n components)
    past_feats = np.expand_dims(past_feats, axis=0).astype(series.dtype)
    future_feats = np.expand_dims(future_feats, axis=0).astype(series.dtype)
    # extract static covariates
    if series.has_static_covariates and model.uses_static_covariates:
        static_feats = np.expand_dims(series.static_covariates_values(), axis=0).astype(
            series.dtype
        )
    return past_feats, future_feats, static_feats



from enum import Enum
from typing import Union

import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.logging import get_logger, raise_log
from darts.utils.ts_utils import series2seq

logger = get_logger(__name__)

# Those freqs can be used to divide Time deltas (the others can't):
DIVISIBLE_FREQS = {"D", "h", "H", "T", "min", "s", "S", "L", "ms", "U", "us", "N", "ns"}
# supported built-in sample weight generators for regression and torch models
SUPPORTED_SAMPLE_WEIGHT = {"linear", "exponential"}


class CovariateType(Enum):
    PAST = "past"
    FUTURE_PAST = "future_past"
    HISTORIC_FUTURE = "historic_future"
    FUTURE = "future"
    NONE = None


def _get_matching_index(ts_target: TimeSeries, ts_covariate: TimeSeries, idx: int):
    """
    Given two overlapping series `ts_target` and `ts_covariate` and an index point `idx` of `ts_target`, returns the
    matching index point in `ts_covariate`, based on the ending times of the two series.
    The indices are starting from the end of the series.

    This function is used to jointly slice target and covariate series in datasets. It supports both datetime and
    integer indexed series.

    Note: this function does not check if the matching index value is in `ts_covariate` or not.
    """
    if ts_target.freq != ts_covariate.freq:
        raise_log(
            ValueError(
                "The dataset contains some target/covariates series pair that have incompatible "
                'time axes (not the same "freq") and thus cannot be matched'
            ),
            logger=logger,
        )

    freq = ts_target.freq

    return idx + _index_diff(
        self=ts_target.end_time(), other=ts_covariate.end_time(), freq=freq
    )


def _index_diff(
    self: Union[pd.Timestamp, int], other: Union[pd.Timestamp, int], freq: pd.offsets
):
    """Returns the difference between two indexes `other` and `self` (`other` - `self`) of frequency `freq`."""
    if isinstance(freq, int):
        return int(other - self)

    elif freq.freqstr in DIVISIBLE_FREQS:
        return int((other - self) / freq)

    # /!\ THIS IS TAKING LINEAR TIME IN THE LENGTH OF THE SERIES
    # it won't scale if the end of target and covariates are far apart and the freq is not in DIVISIBLE_FREQS
    # (Not sure there's a way around it for exotic freqs)
    if other >= self:
        return -1 + len(pd.date_range(start=self, end=other, freq=freq))
    else:
        return 1 - len(pd.date_range(start=other, end=self, freq=freq))


def _process_sample_weight(sample_weight, target_series):
    if sample_weight is None:
        return None

    if target_series is None:
        raise_log(
            ValueError("Must supply target `series` when using `sample_weight`."),
            logger=logger,
        )

    # get sample weights
    if isinstance(sample_weight, str):
        if sample_weight not in SUPPORTED_SAMPLE_WEIGHT:
            raise_log(
                ValueError(
                    f"Invalid `sample_weight` value: `'{sample_weight}'`. "
                    f"If a string, must be one of: {SUPPORTED_SAMPLE_WEIGHT}."
                ),
                logger=logger,
            )
        # create global time weights based on the longest target series
        max_len = max(len(target_i) for target_i in target_series)
        if sample_weight == "linear":
            weights = np.linspace(0, 1, max_len)
        else:  # "exponential"
            time_steps = np.linspace(0, 1, max_len)
            weights = np.exp(-10 * (1 - time_steps))
        weights = np.expand_dims(weights, -1).astype(target_series[0].dtype)

        # create sequence of series for tabularization
        sample_weight = [
            TimeSeries.from_times_and_values(
                times=target_i.time_index,
                values=weights[-len(target_i) :],
            )
            for target_i in target_series
        ]

    sample_weight = series2seq(sample_weight)
    if len(target_series) != len(sample_weight):
        raise_log(
            ValueError(
                "The provided sequence of target `series` must have the same length as "
                "the provided sequence of `sample_weight`."
            ),
            logger=logger,
        )
    return sample_weight
